# The MIT License (MIT)
# Copyright (c) 2019 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import warnings
from typing import Tuple, List, Union, Dict, Any

import numpy as np
import xarray as xr
from osgeo import gdal

from xcube.constants import CRS_WKT_EPSG_4326, GLOBAL_GEO_EXTENT

gdal.UseExceptions()
gdal.PushErrorHandler('CPLQuietErrorHandler')

DEFAULT_RESAMPLING = 'Nearest'
DEFAULT_TP_RESAMPLING = 'Bilinear'

CoordRange = Tuple[float, float, float, float]


def reproject_xy_to_wgs84(src_dataset: xr.Dataset,
                          src_xy_var_names: Tuple[str, str],
                          src_xy_tp_var_names: Tuple[str, str] = None,
                          src_xy_crs: str = None,
                          src_xy_gcp_step: Union[int, Tuple[int, int]] = 10,
                          src_xy_tp_gcp_step: Union[int, Tuple[int, int]] = 1,
                          dst_size: Tuple[int, int] = None,
                          dst_region: CoordRange = None,
                          dst_resampling: Union[str, Dict[str, str]] = DEFAULT_RESAMPLING,
                          include_xy_vars: bool = False,
                          include_non_spatial_vars: bool = False) -> xr.Dataset:
    """
    Reprojection of xarray datasets with 2D geo-coding, e.g. with variables lon(y,x), lat(y, x) to
    EPSG:4326 (WGS-84) coordinate reference system.

    If *dst_resampling* is a string, it provides the default resampling for all variables.
    If *dst_resampling* is a dictionary, it provides a mapping from variable names to the desired
    resampling for that variable.

    The resampling may be one of the following up-sampling algorithms:

    * ``Nearest``
    * ``Bilinear``
    * ``Cubic``
    * ``CubicSpline``
    * ``Lanczos``

    Or one of the down-sampling algorithms:

    * ``Average``
    * ``Min``
    * ``Max``
    * ``Median``
    * ``Mode``
    * ``Q1``
    * ``Q3``

    :param src_dataset:
    :param src_xy_var_names:
    :param src_xy_tp_var_names:
    :param src_xy_crs:
    :param src_xy_gcp_step:
    :param src_xy_tp_gcp_step:
    :param dst_size:
    :param dst_region:
    :param dst_resampling: The spatial resampling algorithm. Either a string that provides the default resampling
           algorithm name or a dictionary that maps variable names to per-variable resampling algorithm names.
    :param include_non_spatial_vars:
    :param include_xy_vars: Whether to include the variables given by *src_xy_var_names*.
           Useful for projection-validation.
    :return: the reprojected dataset
    """
    x_name, y_name = src_xy_var_names
    tp_x_name, tp_y_name = src_xy_tp_var_names or (None, None)

    # Set defaults
    src_xy_crs = src_xy_crs or CRS_WKT_EPSG_4326
    gcp_i_step, gcp_j_step = (src_xy_gcp_step, src_xy_gcp_step) if isinstance(src_xy_gcp_step, int) \
        else src_xy_gcp_step
    tp_gcp_i_step, tp_gcp_j_step = (src_xy_tp_gcp_step, src_xy_tp_gcp_step) if src_xy_tp_gcp_step is None or isinstance(
        src_xy_tp_gcp_step, int) \
        else src_xy_tp_gcp_step

    dst_width, dst_height = dst_size

    _assert(src_dataset is not None)
    _assert(dst_width > 1)
    _assert(dst_height > 1)
    _assert(gcp_i_step > 0)
    _assert(gcp_j_step > 0)

    _assert(x_name in src_dataset)
    _assert(y_name in src_dataset)
    x_var = src_dataset[x_name]
    y_var = src_dataset[y_name]
    if len(x_var.dims) == 1 and len(y_var.dims) == 1:
        y_var, x_var = xr.broadcast(y_var, x_var)
    _assert(len(x_var.dims) == 2)
    _assert(y_var.dims == x_var.dims)
    _assert(x_var.shape[-1] >= 2)
    _assert(x_var.shape[-2] >= 2)
    _assert(y_var.shape == x_var.shape)

    src_width = x_var.shape[-1]
    src_height = x_var.shape[-2]

    dst_region = _ensure_valid_region(dst_region, GLOBAL_GEO_EXTENT, x_var, y_var)
    dst_x1, dst_y1, dst_x2, dst_y2 = dst_region

    dst_res = max((dst_x2 - dst_x1) / dst_width, (dst_y2 - dst_y1) / dst_height)
    _assert(dst_res > 0)

    dst_geo_transform = (dst_x1, dst_res, 0.0,
                         dst_y2, 0.0, -dst_res)

    # Extract GCPs from full-res lon/lat 2D variables
    gcps = _get_gcps(x_var, y_var, gcp_i_step, gcp_j_step)

    if tp_x_name and tp_y_name and tp_x_name in src_dataset and tp_y_name in src_dataset:
        # If there are tie-point variables in the src_dataset
        tp_x_var = src_dataset[tp_x_name]
        tp_y_var = src_dataset[tp_y_name]
        _assert(len(tp_x_var.shape) == 2)
        _assert(tp_x_var.shape == tp_y_var.shape)
        tp_width = tp_x_var.shape[-1]
        tp_height = tp_x_var.shape[-2]
        _assert(tp_gcp_i_step is not None and tp_gcp_i_step > 0)
        _assert(tp_gcp_j_step is not None and tp_gcp_j_step > 0)
        # Extract GCPs also from tie-point lon/lat 2D variables
        tp_gcps = _get_gcps(tp_x_var, tp_y_var, tp_gcp_i_step, tp_gcp_j_step)
    else:
        # No tie-point variables
        tp_x_var = None
        tp_width = None
        tp_height = None
        tp_gcps = None

    mem_driver = gdal.GetDriverByName("MEM")

    dst_x2 = dst_x1 + dst_res * dst_width
    dst_y1 = dst_y2 - dst_res * dst_height

    dst_dataset = _new_dst_dataset(dst_width, dst_height, dst_res, dst_x1, dst_y1, dst_x2, dst_y2)

    if dst_resampling is None:
        dst_resampling = {}
    if isinstance(dst_resampling, str):
        dst_resampling = {var_name: dst_resampling for var_name in src_dataset.variables}

    for var_name in src_dataset.variables:
        src_var = src_dataset[var_name]

        if src_var.dims == x_var.dims:
            is_tp_var = False
            if var_name == x_name or var_name == y_name:
                if not include_xy_vars:
                    # Don't store lat and lon 2D vars in destination
                    continue
                dst_var_name = 'src_' + var_name
            else:
                dst_var_name = var_name
            # PERF: collect variables of same type and size and set band_count accordingly to speed up reprojection
            band_count = 1
            data_type = numpy_to_gdal_dtype(src_var.dtype)
            src_var_dataset = mem_driver.Create(f'src_{var_name}', src_width, src_height, band_count, data_type, [])
            src_var_dataset.SetGCPs(gcps, src_xy_crs)
        elif tp_x_var is not None and src_var.dims == tp_x_var.dims:
            is_tp_var = True
            if var_name == tp_x_name or var_name == tp_y_name:
                if not include_xy_vars:
                    # Don't store lat and lon 2D vars in destination
                    continue
                dst_var_name = 'src_' + var_name
            else:
                dst_var_name = var_name
            # PERF: collect variables of same type and size and set band_count accordingly to speed up reprojection
            band_count = 1
            data_type = numpy_to_gdal_dtype(src_var.dtype)
            src_var_dataset = mem_driver.Create(f'src_{var_name}', tp_width, tp_height, band_count, data_type, [])
            src_var_dataset.SetGCPs(tp_gcps, src_xy_crs)
        elif include_non_spatial_vars:
            # Store any variable as-is, that does not have the lat/lon 2D dims, then continue
            dst_dataset[var_name] = src_var
            continue
        else:
            continue

        # We use GDT_Float64 to introduce NaN as no-data-value
        dst_data_type = gdal.GDT_Float64
        dst_var_dataset = mem_driver.Create(f'dst_{var_name}', dst_width, dst_height, band_count, dst_data_type, [])
        dst_var_dataset.SetProjection(CRS_WKT_EPSG_4326)
        dst_var_dataset.SetGeoTransform(dst_geo_transform)

        # TODO (forman): PERFORMANCE: stack multiple variables of same src_data_type
        #                to perform the reprojection only once per stack

        # TODO (forman): CODE-DUPLICATION: refactor out common code block in reproject_crs_to_wgs84()

        for band_index in range(1, band_count + 1):
            src_var_dataset.GetRasterBand(band_index).SetNoDataValue(float('nan'))
            src_var_dataset.GetRasterBand(band_index).WriteArray(src_var.values)
            dst_var_dataset.GetRasterBand(band_index).SetNoDataValue(float('nan'))

        resample_alg, resample_alg_name = _get_resample_alg(dst_resampling,
                                                            var_name,
                                                            default=DEFAULT_TP_RESAMPLING if is_tp_var else DEFAULT_RESAMPLING)

        warp_mem_limit = 0
        error_threshold = 0
        # See http://www.gdal.org/structGDALWarpOptions.html
        options = ['INIT_DEST=NO_DATA']
        gdal.ReprojectImage(src_var_dataset,
                            dst_var_dataset,
                            None,
                            None,
                            resample_alg,
                            warp_mem_limit,
                            error_threshold,
                            None,  # callback,
                            None,  # callback_data,
                            options)  # options

        dst_values = dst_var_dataset.GetRasterBand(1).ReadAsArray()
        # print(var_name, dst_values.shape, np.nanmin(dst_values), np.nanmax(dst_values))

        dst_dataset[dst_var_name] = _new_dst_variable(src_var, dst_values, resample_alg_name)

    return dst_dataset


def _new_dst_variable(src_var, dst_values, resample_alg_name):
    dst_var_attrs = dict(**src_var.attrs, spatial_resampling=resample_alg_name)
    dst_attrs = dict(**src_var.attrs, spatial_resampling=resample_alg_name)
    if "grid_mapping" in dst_attrs:
        del dst_attrs["grid_mapping"]

    dst_var = xr.DataArray(dst_values, dims=['lat', 'lon'], attrs=dst_var_attrs)
    dst_var.encoding = src_var.encoding
    if np.issubdtype(dst_var.dtype, np.floating) \
            and np.issubdtype(src_var.encoding.get('dtype'), np.integer) \
            and src_var.encoding.get('_FillValue') is None:
        warnings.warn(f'variable {dst_var.name!r}: setting _FillValue=0 to replace any NaNs')
        dst_var.encoding['_FillValue'] = 0
    return dst_var


def _new_dst_dataset(dst_width, dst_height, dst_res, dst_x1, dst_y1, dst_x2, dst_y2):
    return xr.Dataset(coords=dict(
        lon=xr.DataArray(np.linspace(dst_x1 + dst_res / 2, dst_x2 - dst_res / 2, dst_width),
                         dims=['lon', ],
                         attrs=dict(**_LON_ATTRS, bounds='lon_bnds')),
        lat=xr.DataArray(np.linspace(dst_y2 - dst_res / 2, dst_y1 + dst_res / 2, dst_height),
                         dims=['lat', ],
                         attrs=dict(**_LAT_ATTRS, bounds='lat_bnds')),
        lon_bnds=xr.DataArray(list(zip(np.linspace(dst_x1, dst_x2 - dst_res, dst_width),
                                       np.linspace(dst_x1 + dst_res, dst_x2, dst_width))),
                              dims=['lon', 'bnds'],
                              attrs=_LON_ATTRS),
        lat_bnds=xr.DataArray(list(zip(np.linspace(dst_y2, dst_y1 + dst_res, dst_height),
                                       np.linspace(dst_y2 - dst_res, dst_y1, dst_height))),
                              dims=['lat', 'bnds'],
                              attrs=_LAT_ATTRS)
    ))


_LON_ATTRS = dict(long_name='longitude', standard_name='longitude', units='degrees_east')
_LAT_ATTRS = dict(long_name='latitude', standard_name='latitude', units='degrees_north')

_NUMPY_TO_GDAL_DTYPE_MAPPING = {
    np.dtype(np.int8): gdal.GDT_Int16,
    np.dtype(np.int16): gdal.GDT_Int16,
    np.dtype(np.int32): gdal.GDT_Int32,
    np.dtype(np.uint8): gdal.GDT_Byte,
    np.dtype(np.uint16): gdal.GDT_UInt16,
    np.dtype(np.uint32): gdal.GDT_UInt32,
    np.dtype(np.float16): gdal.GDT_Float32,
    np.dtype(np.float32): gdal.GDT_Float32,
    np.dtype(np.float64): gdal.GDT_Float64,
}


def numpy_to_gdal_dtype(np_dtype):
    if np_dtype in _NUMPY_TO_GDAL_DTYPE_MAPPING:
        return _NUMPY_TO_GDAL_DTYPE_MAPPING[np_dtype]
    warnings.warn(f'unhandled numpy dtype {np_dtype}, using float64 instead')
    return gdal.GDT_Float64


def _get_resample_alg(dst_resampling, var_name, default):
    resample_alg_name = dst_resampling.get(var_name, default)
    if resample_alg_name not in NAME_TO_GDAL_RESAMPLE_ALG:
        raise ValueError(f'{resample_alg_name!r} is not a name of a known resampling algorithm')
    resample_alg = NAME_TO_GDAL_RESAMPLE_ALG[resample_alg_name]
    return resample_alg, resample_alg_name


NAME_TO_GDAL_RESAMPLE_ALG: Dict[str, Any] = dict(

    # Up-sampling
    Nearest=gdal.GRA_NearestNeighbour,
    Bilinear=gdal.GRA_Bilinear,
    Cubic=gdal.GRA_Cubic,
    CubicSpline=gdal.GRA_CubicSpline,
    Lanczos=gdal.GRA_Lanczos,

    # Down-sampling
    Average=gdal.GRA_Average,
    Min=gdal.GRA_Min,
    Max=gdal.GRA_Max,
    Median=gdal.GRA_Med,
    Mode=gdal.GRA_Mode,
    Q1=gdal.GRA_Q1,
    Q3=gdal.GRA_Q3,
)


def _assert(cond, text='Assertion failed'):
    if not cond:
        raise ValueError(text)


def _ensure_valid_region(region: CoordRange,
                         valid_region: CoordRange,
                         x_var: xr.DataArray,
                         y_var: xr.DataArray,
                         extra: float = 0.01):
    if region:
        # Extract region
        _assert(len(region) == 4)
        x1, y1, x2, y2 = region
    else:
        # Determine region from full-res lon/lat 2D variables
        x1, x2 = x_var.min(), x_var.max()
        y1, y2 = y_var.min(), y_var.max()
        extra_space = extra * max(x2 - x1, y2 - y1)
        # Add extra space in units of the source coordinates
        x1, x2 = x1 - extra_space, x2 + extra_space
        y1, y2 = y1 - extra_space, y2 + extra_space
    x_min, y_min, x_max, y_max = valid_region
    x1, x2 = max(x_min, x1), min(x_max, x2)
    y1, y2 = max(y_min, y1), min(y_max, y2)
    assert x1 < x2 and y1 < y2
    return x1, y1, x2, y2


def _get_gcps(x_var: xr.DataArray,
              y_var: xr.DataArray,
              i_step: int,
              j_step: int) -> List[gdal.GCP]:
    x_values = x_var.values
    y_values = y_var.values
    i_size = x_var.shape[-1]
    j_size = x_var.shape[-2]
    gcps = []
    gcp_id = 0
    i_count = (i_size + i_step - 1) // i_step
    j_count = (j_size + j_step - 1) // j_step
    for j in np.linspace(0, j_size - 1, j_count, dtype=np.int32):
        for i in np.linspace(0, i_size - 1, i_count, dtype=np.int32):
            x, y = float(x_values[j, i]), float(y_values[j, i])
            gcps.append(gdal.GCP(x, y, 0.0, i + 0.5, j + 0.5, '%s,%s' % (i, j), str(gcp_id)))
            gcp_id += 1
    return gcps


def get_projection_wkt(name: str,
                       proj_name: str,
                       latitude_of_origin: float = 0.0,
                       central_meridian: float = 0.0,
                       scale_factor: float = 1.0,
                       false_easting: float = 0.0,
                       false_northing: float = 0.0):
    return (
        f'PROJCS["{name}",'
        f'  GEOGCS["WGS 84",'
        f'    DATUM["WGS_1984",'
        f'      SPHEROID["WGS 84", 6378137, 298.257223563,'
        f'        AUTHORITY["EPSG", 7030]],'
        f'      TOWGS84[0,0,0,0,0,0,0],'
        f'      AUTHORITY["EPSG", 6326]],'
        f'    PRIMEM["Greenwich", 0, AUTHORITY["EPSG", 8901]],'
        f'    UNIT["DMSH",0.0174532925199433,AUTHORITY["EPSG", 9108]],'
        f'    AXIS["Lat", NORTH],'
        f'    AXIS["Long", EAST],'
        f'    AUTHORITY["EPSG", 4326]],'
        f'  PROJECTION["{proj_name}"],'
        f'  PARAMETER["latitude_of_origin", {latitude_of_origin}],'
        f'  PARAMETER["central_meridian", {central_meridian}],'
        f'  PARAMETER["scale_factor", {scale_factor}],'
        f'  PARAMETER["false_easting", {false_easting}],'
        f'  PARAMETER["false_northing", {false_northing}],'
        f'  UNIT["metre", 1, AUTHORITY["EPSG", "9001"]]'
        f']'
    )



# The MIT License (MIT)
# Copyright (c) 2022 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import abc
from typing import Sequence, List, Type, Optional, Tuple

from xcube.constants import EXTENSION_POINT_SERVER_FRAMEWORKS
from xcube.util.extension import get_extension_registry
from xcube.util.jsonschema import JsonObjectSchema
from .api import ApiRoute
from .api import ApiStaticRoute
from .api import Context
from .asyncexec import AsyncExecution


class Framework(AsyncExecution, abc.ABC):
    """
    An abstract web server framework.
    """

    @property
    @abc.abstractmethod
    def config_schema(self) -> Optional[JsonObjectSchema]:
        """Returns an optional JSON Schema for
        the web server configuration. Returning None
        indicates that configuration is not possible."""

    @abc.abstractmethod
    def add_static_routes(self,
                          routes: Sequence[ApiStaticRoute],
                          url_prefix: str):
        """
        Adds the given static routes to this web server.

        :param routes: The static routes to be added.
        :param url_prefix: URL prefix, may be an empty string.
        """

    @abc.abstractmethod
    def add_routes(self,
                   routes: Sequence[ApiRoute],
                   url_prefix: str):
        """
        Adds the given routes to this web server.

        :param routes: The routes to be added.
        :param url_prefix: URL prefix, may be an empty string.
        """

    @abc.abstractmethod
    def update(self, ctx: Context):
        """
        Called, when the server context has changed.

        This is the case immediately after instantiation and before
        start() is called. It may then be called on any context change,
        likely due to a configuration change.

        :param ctx: The current server context.
        """

    @abc.abstractmethod
    def start(self, ctx: Context):
        """
        Starts the web service.
        :param ctx: The initial server context.
        """

    @abc.abstractmethod
    def stop(self, ctx: Context):
        """
        Stops the web service.
        :param ctx: The current server context.
        """


def get_framework_names() -> List[str]:
    """Get the names of possible web server frameworks."""
    extension_registry = get_extension_registry()
    return [
        ext.name for ext in extension_registry.find_extensions(
            EXTENSION_POINT_SERVER_FRAMEWORKS
        )
    ]


def get_framework_class(framework_name: str) -> Type[Framework]:
    """Get the web server framework class for the given *framework_name*."""
    extension_registry = get_extension_registry()
    return extension_registry.get_component(
        EXTENSION_POINT_SERVER_FRAMEWORKS,
        framework_name
    )



# The MIT License (MIT)
# Copyright (c) 2020 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import json
import sys
import traceback
from typing import Dict, Any

import click

from xcube.cli.common import (cli_option_quiet,
                              cli_option_verbosity,
                              configure_cli_output)
from xcube.constants import LOG


@click.command(name="gen2", hidden=True)
@click.argument('request_path',
                type=str,
                required=False,
                metavar='REQUEST')
@click.option('--stores', 'stores_config_path',
              metavar='STORES_CONFIG',
              help='A JSON or YAML file that maps store names to '
                   'parameterized data stores.')
@click.option('--service', 'service_config_path',
              metavar='SERVICE_CONFIG',
              help='A JSON or YAML file that provides the configuration for an'
                   ' xcube Generator Service. If provided, the REQUEST will be'
                   ' passed to the service instead of generating the cube on '
                   ' this computer.')
@click.option('--info', '-i',
              is_flag=True,
              help='Output information about the data cube to be generated.'
                   ' Does not generate the data cube.')
@click.option('--output', '-o', 'output_file',
              metavar='RESULT',
              help='Write cube information or generation result into JSON'
                   ' file RESULT.'
                   ' If omitted, the JSON is dumped to stdout.')
@cli_option_quiet
@cli_option_verbosity
def gen2(request_path: str,
         stores_config_path: str = None,
         service_config_path: str = None,
         output_file: str = None,
         info: bool = False,
         quiet: bool = False,
         verbosity: int = 0):
    """
    Generator tool for data cubes.

    Creates a cube view from one or more cube stores, optionally performs some
    cube transformation, and writes the resulting cube to some target cube
    store.

    REQUEST is the cube generation request. It may be provided as a JSON or
    YAML file (file extensions ".json" or ".yaml"). If the REQUEST file
    argument is omitted, it is expected that the Cube generation request is
    piped as a JSON string.

    STORE_CONFIGS is a path to a JSON or YAML file with xcube data store
    configurations. It is a mapping of arbitrary store names to configured
    data stores. Entries are dictionaries that have a mandatory "store_id"
    property which is a name of a registered xcube data store.
    The optional "store_params" property may define data store specific
    parameters. The following example defines a data store named
    "my_s3_store" which is an AWS S3 bucket store, and a data store
    "my_test_store" for testing, which is an in-memory data store:

    \b
    {
        "my_s3_store": {
            "store_id": "s3",
            "store_params": {
                "bucket_name": "eurodatacube",
                "aws_access_key_id": "jokljkjoiqqjvlaksd",
                "aws_secret_access_key": "1728349182734983248234"
            }
        },
        "my_test_store": {
            "store_id": "memory"
        }
    }

    SERVICE_CONFIG is a path to a JSON or YAML file with an xcube Generator
    service configuration. Here is a JSON example:

    \b
    {
        "endpoint_url": "https://xcube-gen.eurodatacube.com/api/v2/",
        "client_id": "ALDLPUIOSD5NHS3103",
        "client_secret": "lfaolb3klv904kj23lkdfsjkf430894341"
    }

    Values in the file may also be interpolated from current
    environment variables. In this case templates are used, for example:

    \b
    {
        "endpoint_url": "https://xcube-gen.eurodatacube.com/api/v2/",
        "client_id": "${XCUBE_GEN_CLIENT_ID}",
        "client_secret": "${XCUBE_GEN_CLIENT_SECRET}"
    }

    """
    from xcube.core.gen2 import CubeGenerator
    from xcube.core.gen2 import CubeGeneratorError
    from xcube.core.gen2 import CubeGeneratorRequest
    from xcube.util.versions import get_xcube_versions

    configure_cli_output(quiet=quiet, verbosity=verbosity)

    error = None

    result: Dict[str, Any]

    # noinspection PyBroadException
    try:
        request = CubeGeneratorRequest.from_file(request_path,
                                                 verbosity=verbosity)
        generator = CubeGenerator.new(
            stores_config=stores_config_path,
            service_config=service_config_path,
            verbosity=verbosity
        )

        if info:
            result = generator.get_cube_info(request).to_dict()
        else:
            result = generator.generate_cube(request).to_dict()
    except BaseException as e:
        error = e

        result = dict(status='error',
                      message=f'{error}',
                      traceback=traceback.format_tb(error.__traceback__))
        if isinstance(error, CubeGeneratorError):
            if error.status_code is not None:
                result.update(status_code=error.status_code)
            if error.remote_output is not None:
                result.update(remote_output=error.remote_output)
            if error.remote_traceback is not None:
                result.update(remote_traceback=error.remote_traceback)
        if 'status_code' not in result:
            result['status_code'] = 500

    if result.get('status') == 'error':
        result['versions'] = get_xcube_versions()

    if output_file is not None:
        with open(output_file, 'w') as fp:
            json.dump(result, fp, indent=2)
        LOG.info(f'Result written to {output_file}')
    else:
        json.dump(result, sys.stdout, indent=2)

    if result.get('status') == 'error':
        message = result.get('message',
                             'Cube generation failed.'
                             if error is None else f'{error}')
        click_error = click.ClickException(message)
        raise click_error from error


if __name__ == '__main__':
    gen2()



# The MIT License (MIT)
# Copyright (c) 2023 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.
import json
import re
from typing import Collection, Optional
import fnmatch
from xcube.server.api import ApiHandler, ApiRequest, ApiError
from .api import api
from .context import CoveragesContext
from .controllers import (
    get_coverage_as_json,
    get_coverage_domainset,
    get_coverage_rangetype,
    get_collection_metadata,
    get_coverage_data,
)


PATH_PREFIX = '/ogc'


# noinspection PyAbstractClass,PyMethodMayBeStatic
@api.route(PATH_PREFIX + '/collections/{collectionId}/coverage')
@api.route(PATH_PREFIX + '/collections/{collectionId}/coverage/')
class CoveragesCoverageHandler(ApiHandler[CoveragesContext]):
    """
    Return coverage data

    This is the main coverage endpoint: the one that returns the actual
    data (as opposed to metadata), as TIFF or NetCDF. It can also provide
    representations in HTML and JSON.
    """

    # noinspection PyPep8Naming
    @api.operation(
        operation_id='coveragesCoverage',
        summary='A coverage in OGC API - Coverages',
    )
    async def get(self, collectionId: str):
        ds_ctx = self.ctx.datasets_ctx

        # The single-component type specifiers aren't RFC2045-compliant,
        #  but the OGC API - Coverages draft allows them in the f parameter.
        available_types = [
            'png',
            'image/png',
            'image/tiff',
            'application/x-geotiff',
            'geotiff',
            'application/netcdf',
            'application/x-netcdf',
            'netcdf',
            'html',
            'text/html',
            'json',
            'application/json',
            # TODO: support covjson
        ]
        content_type = negotiate_content_type(self.request, available_types)
        content_bbox = content_crs = None
        if content_type is None:
            raise ApiError.UnsupportedMediaType(
                f'Available media types: {", ".join(available_types)}\n'
            )
        elif content_type in {'text/html', 'html'}:
            result = (
                '<html><title>Collection</title><body><pre>\n'
                + json.dumps(
                    get_coverage_as_json(ds_ctx, collectionId), indent=2
                )
                + '\n</pre></body></html>'
            )
        elif content_type in {'application/json', 'json'}:
            result = get_coverage_as_json(ds_ctx, collectionId)
        else:
            result, content_bbox, content_crs = get_coverage_data(
                ds_ctx, collectionId, self.request.query, content_type
            )

        if content_bbox is not None:
            self.response.set_header(
                'Content-Bbox', ','.join(map(str, content_bbox))
            )
        if content_crs is not None:
            self.response.set_header(
                'Content-Crs', f'[{content_crs.to_string()}]'
            )
        return await self.response.finish(result, content_type=content_type)


# noinspection PyAbstractClass,PyMethodMayBeStatic
@api.route(PATH_PREFIX + '/collections/{collectionId}/coverage/domainset')
@api.route(PATH_PREFIX + '/collections/{collectionId}/coverage/domainset/')
class CoveragesDomainsetHandler(ApiHandler[CoveragesContext]):
    """
    Describe the domain set of a coverage

    The domain set is the set of input parameters (e.g. geographical extent,
    time span) for which the coverage is defined.
    """

    # noinspection PyPep8Naming
    @api.operation(
        operation_id='coveragesDomainSet',
        summary='OGC API - Coverages - domain set',
    )
    async def get(self, collectionId: str):
        domain_set = get_coverage_domainset(
            self.ctx.datasets_ctx, collectionId
        )
        return self.response.finish(domain_set)


# noinspection PyAbstractClass,PyMethodMayBeStatic
@api.route(PATH_PREFIX + '/collections/{collectionId}/coverage/rangetype')
@api.route(PATH_PREFIX + '/collections/{collectionId}/coverage/rangetype/')
class CoveragesRangetypeHandler(ApiHandler[CoveragesContext]):
    """
    Describe the range type of a coverage

    The range type describes the types of the data contained in the range
    of this coverage. For a data cube, these would typically correspond to
    the types of the variables or bands.
    """

    # noinspection PyPep8Naming
    @api.operation(
        operation_id='coveragesRangeType',
        summary='OGC API - Coverages - range type',
    )
    async def get(self, collectionId: str):
        range_type = get_coverage_rangetype(
            self.ctx.datasets_ctx, collectionId
        )
        return self.response.finish(range_type)


@api.route(PATH_PREFIX + '/collections/{collectionId}/coverage/metadata')
@api.route(PATH_PREFIX + '/collections/{collectionId}/coverage/metadata/')
class CoveragesMetadataHandler(ApiHandler[CoveragesContext]):
    """
    Return coverage metadata

    The metadata is taken from the source dataset's attributes
    """

    # noinspection PyPep8Naming
    @api.operation(
        operation_id='coveragesMetadata',
        summary='OGC API - Coverages - metadata',
    )
    async def get(self, collectionId: str):
        return self.response.finish(
            get_collection_metadata(self.ctx.datasets_ctx, collectionId)
        )


@api.route(PATH_PREFIX + '/collections/{collectionId}/coverage/rangeset')
@api.route(PATH_PREFIX + '/collections/{collectionId}/coverage/rangeset/')
class CoveragesRangesetHandler(ApiHandler[CoveragesContext]):
    """
    Handle rangeset endpoint with a "not allowed" response

    This endpoint has been deprecated
    (see https://github.com/m-mohr/geodatacube-api/pull/1 ),
    but we handle it to make clear that its non-implementation is a deliberate
    decision.
    """

    # noinspection PyPep8Naming
    @api.operation(
        operation_id='coveragesRangeset',
        summary='OGC API - Coverages - rangeset',
    )
    async def get(self, collectionId: str):
        self.response.set_status(405)
        self.response.set_header('Allow', '')  # no methods supported
        return self.response.finish(
            'The rangeset endpoint has been deprecated and is not supported.'
        )


def negotiate_content_type(
    request: ApiRequest, available: Collection[str]
) -> Optional[str]:
    """
    Determine a MIME content type based on client and server capabilities

    Client preferences are determined both from the standard HTTP Accept
    header provided by the client, and by an optional `f` parameter which
    can override the Accept header.

    :param request: HTTP request with Accept header and/or f parameter
    :param available: List of MIME types that the server can produce
    :return: the MIME type that is most acceptable to both client and server,
             or None if there is no MIME type acceptable to both
    """
    if 'f' in request.query:  # overrides headers
        content_type = request.query['f'][0]
        return (
            content_type
            if available is None or content_type in available
            else None
        )

    accept = re.split(', *', request.headers.get('Accept'))

    def parse_part(part: str) -> tuple[float, str]:
        if ';q=' in part:
            subparts = part.split(';q=')
            return float(subparts[1]), subparts[0]
        else:
            return 1, part

    type_specs = sorted([parse_part(part) for part in accept], reverse=True)
    types = [ts[1] for ts in type_specs]
    for allowed_type in types:
        for available_type in available:
            # We (ab)use fnmatch to match * wildcards from accept headers
            if fnmatch.fnmatch(available_type, allowed_type):
                return available_type
    return None



# The MIT License (MIT)
# Copyright (c) 2022 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import io
import logging
import math
from typing import Optional, Tuple, Dict, Any, Hashable, Union, Sequence, List

import PIL
import matplotlib.colors
import numpy as np
import pyproj
import xarray as xr

from xcube.constants import LOG
from xcube.util.assertions import assert_in
from xcube.util.assertions import assert_instance
from xcube.util.assertions import assert_true
from xcube.util.cmaps import ColormapProvider
from xcube.util.perf import measure_time_cm
from xcube.util.projcache import ProjCache
from xcube.util.timeindex import ensure_time_label_compatible
from xcube.util.types import Pair
from xcube.util.types import ScalarOrPair
from xcube.util.types import normalize_scalar_or_pair
from .mldataset import MultiLevelDataset
from .tilingscheme import DEFAULT_CRS_NAME
from .tilingscheme import DEFAULT_TILE_SIZE
from .tilingscheme import TilingScheme

DEFAULT_VALUE_RANGE = (0., 1.)
DEFAULT_CMAP_NAME = 'bone'
DEFAULT_FORMAT = 'png'
DEFAULT_TILE_ENLARGEMENT = 1

ValueRange = Tuple[float, float]


def compute_tiles(
        ml_dataset: MultiLevelDataset,
        variable_names: Union[str, Sequence[str]],
        tile_bbox: Tuple[float, float, float, float],
        tile_crs: Union[str, pyproj.CRS] = DEFAULT_CRS_NAME,
        tile_size: ScalarOrPair[int] = DEFAULT_TILE_SIZE,
        level: int = 0,
        non_spatial_labels: Optional[Dict[str, Any]] = None,
        as_dataset: bool = False,
        tile_enlargement: int = DEFAULT_TILE_ENLARGEMENT,
        trace_perf: bool = False,
) -> Optional[Union[List[np.ndarray], xr.Dataset]]:
    """Compute tiles for given *variable_names* in
    given multi-resolution dataset *mr_dataset*.

    The algorithm is as follows:

    1. compute the z-index of the dataset pyramid from requested
       map CRS and tile_z.
    2. select a spatial 2D slice from the dataset pyramid.
    3. create a 2D array of the tile's map coordinates at each pixel
    4. transform 2D array of tile map coordinates into coordinates of
       the dataset CRS.
    5. use the 2D array's outline to retrieve the bbox in dataset
       coordinates.
    6. spatially subset the 2D slice using that bbox.
    7. compute a new bbox from the actual spatial subset.
    8. use the new bbox to transform the 2D array of the tile's dataset
       coordinates into indexes into the spatial subset.
    9. use the tile indices as 2D index into the spatial subset.
       The result is a reprojected array with the shape of the tile.
    10. turn that array into an RGBA image.
    11. encode RGBA image into PNG bytes.

    :param ml_dataset: Multi-level dataset
    :param variable_names: Single variable name
        or a sequence of three names.
    :param tile_bbox: Tile bounding box
    :param tile_crs: Spatial tile coordinate reference system.
        Must be a geographical CRS, such as "EPSG:4326", or
        web mercator, i.e. "EPSG:3857". Defaults to "CRS84".
    :param tile_size: The tile size in pixels.
        Can be a scalar or an integer width/height pair.
        Defaults to 256.
    :param level: Dataset level to be used.
        Defaults to 0, the base level.
    :param non_spatial_labels: Labels for the non-spatial dimensions
        in the given variables.
    :param as_dataset: If set, an ``xr.Dataset`` is returned
        instead of a list of numpy arrays.
    :param tile_enlargement: Enlargement in pixels applied to
        the computed source tile read from the data.
        Can be used to increase the accuracy of the borders of target
        tiles at high zoom levels. Defaults to 1.
    :param trace_perf: If set, detailed performance
        metrics are logged using the level of the "xcube" logger.
    :return: A list of numpy.ndarray instances according to variables
        given by *variable_names*. Returns None, if the resulting
        spatial subset would be too small.
    :raise TileNotFoundException
    """
    if isinstance(variable_names, str):
        variable_names = (variable_names,)

    tile_size = normalize_scalar_or_pair(tile_size)
    tile_width, tile_height = tile_size

    logger = LOG if trace_perf else None

    measure_time = measure_time_cm(disabled=not trace_perf, logger=LOG)

    tile_x_min, tile_y_min, tile_x_max, tile_y_max = tile_bbox

    dataset = ml_dataset.get_dataset(level)

    with measure_time('Preparing 2D subset'):
        variables = [
            _get_variable(ml_dataset.ds_id,
                          dataset,
                          variable_name,
                          non_spatial_labels,
                          logger)
            for variable_name in variable_names
        ]

    variable_0 = variables[0]

    with measure_time('Transforming tile map to dataset coordinates'):
        ds_x_name, ds_y_name = ml_dataset.grid_mapping.xy_dim_names

        ds_y_coords = variable_0[ds_y_name]
        ds_y_points_up = bool(ds_y_coords[0] < ds_y_coords[-1])

        tile_res_x = (tile_x_max - tile_x_min) / (tile_width - 1)
        tile_res_y = (tile_y_max - tile_y_min) / (tile_height - 1)

        tile_x_1d = np.linspace(tile_x_min + 0.5 * tile_res_x,
                                tile_x_max - 0.5 * tile_res_x, tile_width)
        tile_y_1d = np.linspace(tile_y_min + 0.5 * tile_res_y,
                                tile_y_max - 0.5 * tile_res_y, tile_height)

        tile_x_2d = np.tile(tile_x_1d, (tile_height, 1))
        tile_y_2d = np.tile(tile_y_1d, (tile_width, 1)).transpose()

        assert tile_x_2d.shape == (tile_height, tile_width)
        assert tile_y_2d.shape == tile_x_2d.shape

        t_map_to_ds = ProjCache.INSTANCE.get_transformer(
            tile_crs,
            ml_dataset.grid_mapping.crs
        )

        tile_ds_x_2d, tile_ds_y_2d = t_map_to_ds.transform(tile_x_2d,
                                                           tile_y_2d)

    with measure_time('Getting spatial subset'):
        # Get min/max of the 1D arrays surrounding the 2D array
        # North
        ds_x_n = tile_ds_x_2d[0, :]
        ds_y_n = tile_ds_y_2d[0, :]
        # South
        ds_x_s = tile_ds_x_2d[tile_height - 1, :]
        ds_y_s = tile_ds_y_2d[tile_height - 1, :]
        # West
        ds_x_w = tile_ds_x_2d[:, 0]
        ds_y_w = tile_ds_y_2d[:, 0]
        # East
        ds_x_e = tile_ds_x_2d[:, tile_width - 1]
        ds_y_e = tile_ds_y_2d[:, tile_width - 1]
        # Min
        ds_x_min = np.nanmin([np.nanmin(ds_x_n), np.nanmin(ds_x_s),
                              np.nanmin(ds_x_w), np.nanmin(ds_x_e)])
        ds_y_min = np.nanmin([np.nanmin(ds_y_n), np.nanmin(ds_y_s),
                              np.nanmin(ds_y_w), np.nanmin(ds_y_e)])
        # Max
        ds_x_max = np.nanmax([np.nanmax(ds_x_n), np.nanmax(ds_x_s),
                              np.nanmax(ds_x_w), np.nanmax(ds_x_e)])
        ds_y_max = np.nanmax([np.nanmax(ds_y_n), np.nanmax(ds_y_s),
                              np.nanmax(ds_y_w), np.nanmax(ds_y_e)])
        if np.isnan(ds_x_min) or np.isnan(ds_y_min) \
                or np.isnan(ds_y_max) or np.isnan(ds_y_max):
            raise TileNotFoundException(
                'Tile bounds NaN after map projection',
                logger=logger
            )

        num_extra_pixels = tile_enlargement
        res_x = (ds_x_max - ds_x_min) / tile_width
        res_y = (ds_y_max - ds_y_min) / tile_height
        extra_dx = num_extra_pixels * res_x
        extra_dy = num_extra_pixels * res_y
        ds_x_slice = slice(ds_x_min - extra_dx, ds_x_max + extra_dx)
        if ds_y_points_up:
            ds_y_slice = slice(ds_y_min - extra_dy, ds_y_max + extra_dy)
        else:
            ds_y_slice = slice(ds_y_max + extra_dy, ds_y_min - extra_dy)

        var_subsets = [variable.sel({ds_x_name: ds_x_slice,
                                     ds_y_name: ds_y_slice})
                       for variable in variables]
        for var_subset in var_subsets:
            # A zero or a one in the tile's shape will produce a
            # non-existing or too small tile. It will also prevent
            # determining the current resolution.
            if 0 in var_subset.shape or 1 in var_subset.shape:
                return None

    with measure_time('Transforming dataset coordinates into indices'):
        var_subset_0 = var_subsets[0]
        ds_x_coords = var_subset_0[ds_x_name]
        ds_y_coords = var_subset_0[ds_y_name]

        ds_x1 = float(ds_x_coords[0])
        ds_x2 = float(ds_x_coords[-1])
        ds_y1 = float(ds_y_coords[0])
        ds_y2 = float(ds_y_coords[-1])

        ds_size_x = ds_x_coords.size
        ds_size_y = ds_y_coords.size

        ds_dx = (ds_x2 - ds_x1) / (ds_size_x - 1)
        ds_dy = (ds_y2 - ds_y1) / (ds_size_y - 1)

        ds_x_indices = (tile_ds_x_2d - ds_x1) / ds_dx
        ds_y_indices = (tile_ds_y_2d - ds_y1) / ds_dy

        ds_x_indices = ds_x_indices.astype(dtype=np.int64)
        ds_y_indices = ds_y_indices.astype(dtype=np.int64)

    with measure_time('Masking dataset indices'):
        ds_mask = (ds_x_indices >= 0) & (ds_x_indices < ds_size_x) \
                  & (ds_y_indices >= 0) & (ds_y_indices < ds_size_y)

        ds_x_indices = np.where(ds_mask, ds_x_indices, 0)
        ds_y_indices = np.where(ds_mask, ds_y_indices, 0)

    var_tiles = []
    for var_subset in var_subsets:
        with measure_time('Loading 2D data for spatial subset'):
            # Note, we need to load the values here into a numpy array,
            # because 2D indexing by [ds_y_indices, ds_x_indices]
            # does not (yet) work with dask arrays.
            var_tile = var_subset.values
            # Remove any axes above the 2nd. This is safe,
            # they will be of size one, if any.
            var_tile = var_tile.reshape(var_tile.shape[-2:])

        with measure_time('Looking up dataset indices'):
            # This does the actual projection trick.
            # Lookup indices ds_y_indices, ds_x_indices to create
            # the actual tile.
            var_tile = var_tile[ds_y_indices, ds_x_indices]
            var_tile = np.where(ds_mask, var_tile, np.nan)

        var_tiles.append(var_tile)

    if as_dataset:
        return _new_tile_dataset(
            [(var, dataset[var.name].dims) for var in variables],
            var_tiles,
            (ds_x_name, ds_y_name),
            (tile_x_1d, tile_y_1d),
            tile_crs
        )

    return var_tiles


def _new_tile_dataset(
        original_vars: List[Tuple[xr.DataArray, Tuple[Hashable, ...]]],
        tiles: List[np.ndarray],
        xy_names: Tuple[str, str],
        xy_coords: Tuple[np.ndarray, np.ndarray],
        crs: Union[str, pyproj.CRS]):
    data_vars = {}
    non_spatial_coords = {}
    for i, (original_var, original_dims) in enumerate(original_vars):
        var_name = original_var.name
        non_spatial_dims = []
        for dim in original_dims:
            if dim not in xy_names:
                non_spatial_dims.append(dim)
                if dim not in non_spatial_coords \
                        and dim in original_var.coords:
                    non_spatial_coords[dim] = original_var.coords[dim]
        data_2d = tiles[i]
        data_nd = data_2d[(*(len(non_spatial_dims) * [np.newaxis]), ...)]
        data_vars[var_name] = xr.DataArray(
            data=data_nd,
            dims=(*non_spatial_dims, "y", "x"),
            name=var_name,
            attrs=dict(**original_var.attrs, grid_mapping="crs"),
        )
    print(pyproj.CRS(crs).to_cf())
    return xr.Dataset(
        data_vars=dict(
            **data_vars,
            crs=xr.DataArray((), attrs=pyproj.CRS(crs).to_cf())
        ),
        coords=dict(
            **{k: xr.DataArray([v.values], dims=k, attrs=v.attrs)
               for k, v in non_spatial_coords.items()},
            y=xr.DataArray(xy_coords[1], dims="y", attrs=dict(
                long_name="y coordinate of projection",
                standard_name="projection_y_coordinate"
            )),
            x=xr.DataArray(xy_coords[0], dims="x", attrs=dict(
                long_name="x coordinate of projection",
                standard_name="projection_x_coordinate"
            )),
        )
    )


def compute_rgba_tile(
        ml_dataset: MultiLevelDataset,
        variable_names: Union[str, Sequence[str]],
        tile_x: int,
        tile_y: int,
        tile_z: int,
        cmap_provider: ColormapProvider,
        crs_name: str = DEFAULT_CRS_NAME,
        tile_size: ScalarOrPair[int] = DEFAULT_TILE_SIZE,
        cmap_name: str = None,
        value_ranges: Optional[Union[ValueRange,
                                     Sequence[ValueRange]]] = None,
        non_spatial_labels: Optional[Dict[str, Any]] = None,
        format: str = DEFAULT_FORMAT,
        tile_enlargement: int = DEFAULT_TILE_ENLARGEMENT,
        trace_perf: bool = False,
) -> Union[bytes, np.ndarray]:
    """Compute an RGBA image tile from *variable_names* in
    given multi-resolution dataset *mr_dataset*.

    If length of *variable_names* is three, we create a direct component
    RGBA image where the three variables correspond to the three R, G, B
    channels and A is computed from NaN values in the three variables.
    *value_ranges* must have exactly three elements too.

    If length of *variable_names* is one, we create a color mapped
    RGBA image using *cmap_name* where R, G, B is computed from a color bar
    and A is computed from NaN values in the variable.
    *value_ranges* must have exactly one element.

    The algorithm is as follows:

    1. compute the z-index of the dataset pyramid from requested
       map CRS and tile_z.
    2. select a spatial 2D slice from the dataset pyramid.
    3. create a 2D array of the tile's map coordinates at each pixel
    4. transform 2D array of tile map coordinates into coordinates of
       the dataset CRS.
    5. use the 2D array's outline to retrieve the bbox in dataset
       coordinates.
    6. spatially subset the 2D slice using that bbox.
    7. compute a new bbox from the actual spatial subset.
    8. use the new bbox to transform the 2D array of the tile's dataset
       coordinates into indexes into the spatial subset.
    9. use the tile indices as 2D index into the spatial subset.
       The result is a reprojected array with the shape of the tile.
    10. turn that array into an RGBA image.
    11. encode RGBA image into PNG bytes.

    :param ml_dataset: Multi-level dataset
    :param variable_names: Single variable name
        or a sequence of three names.
    :param tile_x: Tile X coordinate
    :param tile_y: Tile Y coordinate
    :param tile_z:  Tile Z coordinate
    :param cmap_provider: Provider for colormaps.
    :param crs_name: Spatial tile coordinate reference system.
        Must be a geographical CRS, such as "EPSG:4326", or
        web mercator, i.e. "EPSG:3857". Defaults to "CRS84".
    :param tile_size: The tile size in pixels.
        Can be a scalar or an integer width/height pair.
        Defaults to 256.
    :param cmap_name: Color map name. Only used if a single
        variable name is given. Defaults to "bone".
    :param value_ranges: A single value range, or value ranges
        for each variable name.
    :param non_spatial_labels: Labels for the non-spatial dimensions
        in the given variables.
    :param tile_enlargement: Enlargement in pixels applied to
        the computed source tile read from the data.
        Can be used to increase the accuracy of the borders of target
        tiles at high zoom levels. Defaults to 1.
    :param format: Either 'png', 'image/png' or 'numpy'.
    :param trace_perf: If set, detailed performance
        metrics are logged using the level of the "xcube" logger.
    :return: PNG bytes or unit8 numpy array, depending on *format*
    :raise TileNotFoundException
    :raise TileRequestException
    """
    if isinstance(variable_names, str):
        variable_names = (variable_names,)
    num_components = len(variable_names)
    assert_true(num_components in (1, 3),
                message='number of names in'
                        ' variable_names must be 1 or 3')
    tile_size = normalize_scalar_or_pair(tile_size)
    tile_width, tile_height = tile_size
    if not value_ranges:
        value_ranges = num_components * (DEFAULT_VALUE_RANGE,)
    else:
        assert_instance(value_ranges, (list, tuple), name='value_ranges')
    if isinstance(value_ranges[0], (int, float)):
        value_ranges = num_components * (value_ranges,)
    assert_true(num_components == len(value_ranges),
                message='value_ranges must have'
                        ' same length as variable_names')
    format = _normalize_format(format)
    assert_in(format, ('png', 'numpy'), name='format')

    measure_time = measure_time_cm(disabled=not trace_perf, logger=LOG)

    tiling_scheme = TilingScheme.for_crs(crs_name) \
        .derive(tile_size=tile_size)

    tile_bbox = tiling_scheme.get_tile_extent(tile_x, tile_y, tile_z)
    if tile_bbox is None:
        # Whether to raise or not
        # could be made a configuration option.
        #
        # raise TileRequestException.new(
        #     'Tile indices out of tile grid bounds',
        #     logger=logger
        # )
        return TransparentRgbaTilePool.INSTANCE.get(tile_size, format)

    ds_level = tiling_scheme.get_resolutions_level(
        tile_z,
        ml_dataset.avg_resolutions,
        ml_dataset.grid_mapping.spatial_unit_name
    )

    var_tiles = compute_tiles(ml_dataset,
                              variable_names,
                              tile_bbox,
                              tiling_scheme.crs,
                              tile_size=tile_size,
                              level=ds_level,
                              non_spatial_labels=non_spatial_labels,
                              tile_enlargement=tile_enlargement,
                              trace_perf=trace_perf)

    if var_tiles is None:
        return TransparentRgbaTilePool.INSTANCE.get(tile_size, format)

    norm_var_tiles = []
    for var_tile, value_range in zip(var_tiles, value_ranges):
        with measure_time('Normalizing data tile'):
            var_tile = var_tile[::-1, :]
            value_min, value_max = value_range
            if value_max < value_min:
                value_min, value_max = value_max, value_min
            if math.isclose(value_min, value_max):
                value_max = value_min + 1
            norm = matplotlib.colors.Normalize(
                value_min, value_max, clip=True
            )
            norm_var_tile = norm(var_tile)

        norm_var_tiles.append(norm_var_tile)

    with measure_time('Encoding tile as RGBA image'):
        if len(norm_var_tiles) == 1:
            var_tile_norm = norm_var_tiles[0]
            _, cm = cmap_provider.get_cmap(cmap_name)
            var_tile_rgba = cm(var_tile_norm)
            var_tile_rgba = (255 * var_tile_rgba).astype(np.uint8)
        else:
            r, g, b = norm_var_tiles
            var_tile_rgba = np.zeros((tile_height, tile_width, 4),
                                     dtype=np.uint8)
            var_tile_rgba[..., 0] = 255 * r
            var_tile_rgba[..., 1] = 255 * g
            var_tile_rgba[..., 2] = 255 * b
            var_tile_rgba[..., 3] = np.where(np.isfinite(r + g + b), 255, 0)

    if format == 'png':
        with measure_time('Encoding RGBA image as PNG bytes'):
            return _encode_rgba_as_png(var_tile_rgba)
    else:  # format == 'numpy'
        return var_tile_rgba


def get_var_cmap_params(var: xr.DataArray,
                        cmap_name: Optional[str],
                        cmap_range: Tuple[Optional[float], Optional[float]],
                        valid_range: Optional[Tuple[float, float]]) \
        -> Tuple[str, Tuple[float, float]]:
    if cmap_name is None:
        cmap_name = var.attrs.get('color_bar_name')
        if cmap_name is None:
            cmap_name = DEFAULT_CMAP_NAME
    cmap_vmin, cmap_vmax = cmap_range
    if cmap_vmin is None:
        cmap_vmin = var.attrs.get('color_value_min')
        if cmap_vmin is None and valid_range is not None:
            cmap_vmin = valid_range[0]
        if cmap_vmin is None:
            cmap_vmin = DEFAULT_VALUE_RANGE[0]
    if cmap_vmax is None:
        cmap_vmax = var.attrs.get('color_value_max')
        if cmap_vmax is None and valid_range is not None:
            cmap_vmax = valid_range[1]
        if cmap_vmax is None:
            cmap_vmax = DEFAULT_VALUE_RANGE[1]
    return cmap_name, (cmap_vmin, cmap_vmax)


def get_var_valid_range(var: xr.DataArray) -> Optional[Tuple[float, float]]:
    valid_min = None
    valid_max = None
    valid_range = var.attrs.get('valid_range')
    if valid_range:
        try:
            valid_min, valid_max = map(float, valid_range)
        except (TypeError, ValueError):
            pass
    if valid_min is None:
        valid_min = var.attrs.get('valid_min')
    if valid_max is None:
        valid_max = var.attrs.get('valid_max')
    if valid_min is None and valid_max is None:
        valid_range = None
    elif valid_min is not None and valid_max is not None:
        valid_range = valid_min, valid_max
    elif valid_min is None:
        valid_range = -np.inf, valid_max
    else:
        valid_range = valid_min, +np.inf
    return valid_range


def _get_variable(ds_name,
                  dataset,
                  variable_name,
                  non_spatial_labels,
                  logger):
    if variable_name not in dataset:
        raise TileNotFoundException(
            f'Variable {variable_name!r}'
            f' not found in dataset {ds_name!r}',
            logger=logger
        )
    variable = dataset[variable_name]
    non_spatial_labels = _get_non_spatial_labels(dataset,
                                                 variable,
                                                 non_spatial_labels,
                                                 logger)
    if non_spatial_labels:
        non_spatial_labels_safe = \
            ensure_time_label_compatible(variable, non_spatial_labels)
        variable = variable.sel(**non_spatial_labels_safe, method='nearest')
    return variable


class TileException(Exception):
    def __init__(self,
                 message: str,
                 logger: Optional[logging.Logger] = None):
        super().__init__(message)
        if logger is not None:
            logger.warning(message)


class TileNotFoundException(TileException):
    pass


class TileRequestException(TileException):
    pass


class TransparentRgbaTilePool:
    """A cache for fully-transparent RGBA tiles of a given size and format."""

    INSTANCE: 'TransparentRgbaTilePool'

    def __init__(self):
        self._transparent_tiles: Dict[str, Union[bytes, np.ndarray]] = dict()

    def get(self, tile_size: Pair[int], format: str) \
            -> Union[bytes, np.ndarray]:
        tile_w, tile_h = tile_size
        key = f'{format}-{tile_w}-{tile_h}'
        if key not in self._transparent_tiles:
            data = np.zeros((tile_h, tile_w, 4), dtype=np.uint8)
            if format == 'png':
                data = _encode_rgba_as_png(data)
            self._transparent_tiles[key] = data
        return self._transparent_tiles[key]


TransparentRgbaTilePool.INSTANCE = TransparentRgbaTilePool()


def _get_non_spatial_labels(dataset: xr.Dataset,
                            variable: xr.DataArray,
                            labels: Optional[Dict[str, Any]],
                            logger: logging.Logger) -> Dict[Hashable, Any]:
    labels = labels if labels is not None else {}

    new_labels = {}
    # assuming last two dims are spatial: [..., y, x]
    assert variable.ndim >= 2
    non_spatial_dims = variable.dims[0:-2]
    if not non_spatial_dims:
        #  Ignore any extra labels passed.
        return new_labels

    for dim in non_spatial_dims:

        try:
            coord_var = dataset.coords[dim]
            if coord_var.size == 0:
                continue
        except KeyError:
            continue

        dim_name = str(dim)

        label = labels.get(dim_name)
        if label is None:
            if logger:
                logger.debug((f'missing label for dimension {dim!r},'
                              f' using first label instead'))
            label = coord_var[0].values

        elif isinstance(label, str):
            if '/' in label:
                # In case of WMTS tile requests the tame range labels
                # from WMTS dimensions may be passed.
                label = label.split('/', maxsplit=1)[0]

            if label.lower() == 'first':
                label = coord_var[0].values
            elif label.lower() in ('last', 'current'):
                label = coord_var[-1].values
            else:
                try:
                    label = np.array(label).astype(coord_var.dtype)
                except (TypeError, ValueError) as e:
                    raise TileRequestException(
                        f'Illegal label {label!r} for dimension {dim!r}'
                    ) from e

        new_labels[dim] = label

    return new_labels


def _normalize_format(format: str) -> str:
    if format in ('png', 'PNG', 'image/png'):
        return 'png'
    return format


def _encode_rgba_as_png(rgba_array: np.ndarray) -> bytes:
    # noinspection PyUnresolvedReferences
    image = PIL.Image.fromarray(rgba_array)
    stream = io.BytesIO()
    image.save(stream, format='PNG')
    return bytes(stream.getvalue())



# The MIT License (MIT)
# Copyright (c) 2022 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Optional, Tuple

import click

from xcube.cli.common import (configure_cli_output,
                              cli_option_quiet,
                              cli_option_verbosity)
from xcube.constants import LOG

DEFAULT_TILE_SIZE = 512
DEFAULT_AGG_METHOD = 'first'


# noinspection PyShadowingBuiltins
@click.command(name="level")
@click.argument('input')
@click.option('--output', '-o', metavar='OUTPUT',
              help='Output path. If omitted, "INPUT.levels" will be used.'
                   ' You can also use S3 object storage URLs of the form'
                   ' "s3://<bucket>/<path>" or "https://<endpoint>"')
@click.option('--link', '-L', is_flag=True, flag_value=True,
              help='Link the INPUT instead of converting it to a level'
                   ' zero dataset. Use with care, as the INPUT\'s internal'
                   ' spatial chunk sizes may be inappropriate'
                   ' for imaging purposes.')
@click.option('--tile-size', '-t', metavar='TILE_SIZE',
              help=f'Tile size, given as single integer number or'
                   f' as <tile-width>,<tile-height>.'
                   f' If omitted, the tile size will be derived'
                   f' from the INPUT\'s'
                   f' internal spatial chunk sizes.'
                   f' If the INPUT is not chunked,'
                   f' tile size will be {DEFAULT_TILE_SIZE}.')
@click.option('--num-levels-max', '-n', metavar='NUM_LEVELS_MAX', type=int,
              help=f'Maximum number of levels to generate.'
                   f' If not given, the number of levels will'
                   f' be derived from spatial dimension and tile sizes.')
@click.option('--agg-methods', '-A', metavar='AGG_METHODS',
              default=DEFAULT_AGG_METHOD,
              help=f'Aggregation method(s) to be used for data variables.'
                   f' Either one of "first", "min", "max", "mean", "median",'
                   f' "auto" or list of assignments to individual variables'
                   f' using the notation'
                   f' "<var1>=<method1>,<var2>=<method2>,..."'
                   f' Defaults to "{DEFAULT_AGG_METHOD}".')
@click.option('--replace', '-r', is_flag=True, flag_value=True,
              help=f'Whether to replace an existing dataset at OUTPUT.')
@click.option('--anon', '-a', is_flag=True, flag_value=True,
              help=f'For S3 inputs or outputs, whether the access'
                   f' is anonymous. By default, credentials are required.')
@cli_option_quiet
@cli_option_verbosity
def level(input: str,
          output: Optional[str],
          link: bool,
          tile_size: Optional[str],
          num_levels_max: int,
          agg_methods: str,
          replace: bool,
          anon: bool,
          quiet: bool,
          verbosity: int):
    """
    Generate multi-resolution levels.

    Transform the given dataset by INPUT into the levels of a
    multi-level pyramid with spatial resolution decreasing by a
    factor of two in both spatial dimensions and write the
    result to directory OUTPUT.

    INPUT may be an S3 object storage URL of the form
    "s3://<bucket>/<path>" or "https://<endpoint>".
    """
    import time
    import os
    from xcube.cli.common import parse_cli_sequence
    from xcube.cli.common import assert_positive_int_item
    from xcube.core.store import new_fs_data_store
    from xcube.core.subsampling import assert_valid_agg_methods

    configure_cli_output(quiet=quiet, verbosity=verbosity)

    input_path = input
    output_path = output
    link_input = link

    if tile_size is not None:
        tile_size = parse_cli_sequence(
            tile_size,
            metavar='TILE_SIZE', num_items=2,
            item_plural_name='tile sizes',
            item_parser=int,
            item_validator=assert_positive_int_item
        )

    if num_levels_max is not None and num_levels_max < 1:
        raise click.ClickException(
            f"NUM_LEVELS_MAX must be a positive integer"
        )

    try:
        if '=' in agg_methods:
            agg_methods = {
                p[0].strip(): (p[1].strip() if len(p) == 2 else None)
                for p in (c.split('=', maxsplit=2)
                          for c in agg_methods.split(','))
            }
        assert_valid_agg_methods(agg_methods)
    except (TypeError, ValueError, SyntaxError) as e:
        raise click.ClickException('invalid AGG_METHODS') from e

    input_protocol, input_path = _split_protocol_and_path(input_path)

    if not output_path:
        dir_path = os.path.dirname(input_path).replace("\\", "/")
        basename, ext = os.path.splitext(os.path.basename(input_path))
        output_protocol = input_protocol
        output_path = f'{dir_path}/{basename + ".levels"}'
    else:
        output_protocol, output_path = _split_protocol_and_path(output_path)

    if link_input and input_protocol != output_protocol:
        raise click.ClickException('Links can be used only'
                                   ' if input and output are in'
                                   ' the same filesystem')

    start_time = time.time()

    input_store = new_fs_data_store(input_protocol,
                                    storage_options=dict(anon=anon)
                                    if input_protocol == 's3' else None)
    input_dataset = input_store.open_data(input_path)

    output_store = new_fs_data_store(output_protocol,
                                     storage_options=dict(anon=anon)
                                     if output_protocol == 's3' else None)
    try:
        output_store.write_data(
            input_dataset,
            output_path,
            replace=replace,
            tile_size=tile_size,
            num_levels=num_levels_max,
            base_dataset_id=input_path if link_input else None,
            agg_methods=agg_methods
        )
    except FileExistsError as e:
        raise click.ClickException(
            f"output {output_path!r} already exists"
        ) from e

    LOG.info(f"Multi-level dataset written to {output_path}"
             f" after {time.time() - start_time} seconds")


def _split_protocol_and_path(path) -> Tuple[str, str]:
    if '://' in path:
        protocol, path = path.split('://', 2)
        if protocol == 'https':
            protocol = 's3'
    else:
        protocol = 'file'
    return protocol, path



# The MIT License (MIT)
# Copyright (c) 2022 by the xcube development team and contributors
# 
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import collections.abc
from typing import Iterator, List

import zarr.storage


class DiagnosticZarrStore(zarr.storage.Store):
    """A diagnostic Zarr store used for testing and investigating
    behaviour of Zarr and xarray's Zarr backend.

    :param store: Wrapped Zarr store.
    """

    def __init__(self, store: collections.abc.MutableMapping):
        self.store = store
        self.records: List[str] = []
        if hasattr(self.store, "listdir"):
            self.listdir = self._listdir
        if hasattr(self.store, "getsize"):
            self.getsize = self._getsize

    def _add_record(self, record: str):
        self.records.append(record)

    def _listdir(self, path: str = "") -> List[str]:
        self._add_record(f"listdir({path!r})")
        # noinspection PyUnresolvedReferences
        return self.store.listdir(path=path)

    def _getsize(self, key: str) -> int:
        self._add_record(f"getsize({key!r})")
        # noinspection PyUnresolvedReferences
        return self.store.getsize(key)

    def keys(self):
        self._add_record("keys()")
        return self.store.keys()

    def __setitem__(self, key: str, value: bytes) -> None:
        self._add_record(f"__setitem__({key!r}, {type(value).__name__})")
        self.store.__setitem__(key, value)

    def __delitem__(self, key: str) -> None:
        self._add_record(f"__delitem__({key!r})")
        self.store.__delitem__(key)

    def __getitem__(self, key: str) -> bytes:
        self._add_record(f"__getitem__({key!r})")
        return self.store.__getitem__(key)

    def __len__(self) -> int:
        self._add_record("__len__()")
        return self.store.__len__()

    def __iter__(self) -> Iterator[str]:
        self._add_record("__iter__()")
        return self.store.__iter__()



# The MIT License (MIT)
# Copyright (c) 2022 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

from typing import Any, Dict, Optional

import xarray as xr

from xcube.core.gridmapping import GridMapping
from xcube.core.schema import rechunk_cube
from xcube.core.subsampling import AggMethods
from xcube.core.subsampling import get_dataset_agg_methods
from xcube.core.subsampling import subsample_dataset
from xcube.core.tilingscheme import get_num_levels
from xcube.util.assertions import assert_instance
from xcube.util.assertions import assert_true
from .lazy import LazyMultiLevelDataset


class BaseMultiLevelDataset(LazyMultiLevelDataset):
    """A multi-level dataset whose level datasets are
    created by down-sampling a base dataset.

    :param base_dataset: The base dataset for the level at index zero.
    :param grid_mapping: Optional grid mapping for *base_dataset*.
    :param num_levels: Optional number of levels.
    :param ds_id: Optional dataset identifier.
    :param agg_methods: Optional aggregation methods.
        May be given as string or as mapping from variable name pattern
        to aggregation method. Valid aggregation methods are
        None, "first", "min", "max", "mean", "median".
        If None, the default, "first" is used for integer variables
        and "mean" for floating point variables.
    """

    def __init__(self,
                 base_dataset: xr.Dataset,
                 grid_mapping: Optional[GridMapping] = None,
                 num_levels: Optional[int] = None,
                 agg_methods: Optional[AggMethods] = 'first',
                 ds_id: Optional[str] = None):
        assert_instance(base_dataset, xr.Dataset, name='base_dataset')
        if grid_mapping is not None:
            assert_instance(grid_mapping, GridMapping, name='grid_mapping')

        if grid_mapping is None:
            # TODO (forman): why not computing it lazily?
            grid_mapping = GridMapping.from_dataset(base_dataset,
                                                    tolerance=1e-4)

        self._agg_methods = get_dataset_agg_methods(
            base_dataset,
            xy_dim_names=grid_mapping.xy_dim_names,
            agg_methods=agg_methods
        )

        self._base_dataset = base_dataset
        super().__init__(grid_mapping=grid_mapping,
                         num_levels=num_levels,
                         ds_id=ds_id)

    @property
    def agg_methods(self):
        return self._agg_methods

    def _get_num_levels_lazily(self) -> int:
        gm = self.grid_mapping
        return get_num_levels(gm.size, gm.tile_size)

    def _get_dataset_lazily(self,
                            index: int,
                            parameters: Dict[str, Any]) -> xr.Dataset:
        """
        Compute the dataset at level *index*: If *index* is zero, return
        the base image passed to constructor, otherwise down-sample the
        dataset for the level at given *index*.

        :param index: the level index
        :param parameters: currently unused
        :return: the dataset for the level at *index*.
        """
        assert_instance(index, int, name='index')
        if index < 0:
            index = self.num_levels + index
        assert_true(0 <= index < self.num_levels,
                    message='index out of range')

        if index == 0:
            level_dataset = self._base_dataset
        else:
            level_dataset = subsample_dataset(
                self._base_dataset,
                2 ** index,
                xy_dim_names=self.grid_mapping.xy_dim_names,
                agg_methods=self._agg_methods
            )

        # Tile each level according to grid mapping
        tile_size = self.grid_mapping.tile_size
        if tile_size is not None:
            level_dataset, _ = rechunk_cube(level_dataset,
                                            self.grid_mapping,
                                            tile_size=tile_size)
        return level_dataset



import unittest

from test.mixins import AlmostEqualDeepMixin


class AlmostEqualDeepMixinTest(unittest.TestCase, AlmostEqualDeepMixin):
    def test_int_and_float_7_places_default(self):
        self.assertAlmostEqualDeep(0, 0.8e-8)
        with self.assertRaises(AssertionError):
            self.assertAlmostEqualDeep(0, 0.8e-7)

    def test_int(self):
        self.assertAlmostEqualDeep(45, 45)
        with self.assertRaises(AssertionError):
            self.assertAlmostEqualDeep(45, 54)

    def test_str(self):
        self.assertAlmostEqualDeep("abc", "abc")
        with self.assertRaises(AssertionError):
            self.assertAlmostEqualDeep("abc", "Abc")

    def test_bool(self):
        self.assertAlmostEqualDeep(True, True)
        with self.assertRaises(AssertionError):
            self.assertAlmostEqualDeep(True, False)

    def test_set(self):
        expected = {'a', 1.1256, True}
        self.assertAlmostEqualDeep(expected, expected)
        self.assertAlmostEqualDeep(expected, {'a', 1.1256, True})
        with self.assertRaises(AssertionError):
            # We currently don't test sets
            self.assertAlmostEqualDeep(expected, {'a', 1.1251, True}, places=2)
        with self.assertRaises(AssertionError):
            self.assertAlmostEqualDeep(expected, {'a', 1.1256, False})

    def test_dict(self):
        expected = {'a': 1.1256, 'b': 5}
        self.assertAlmostEqualDeep(expected, expected)
        self.assertAlmostEqualDeep(expected, {'a': 1.1256, 'b': 5})
        self.assertAlmostEqualDeep(expected, {'a': 1.1251, 'b': 5}, places=3)
        with self.assertRaises(AssertionError):
            self.assertAlmostEqualDeep(expected, {'a': 1.1251, 'b': 5}, places=4)
        with self.assertRaises(AssertionError):
            self.assertAlmostEqualDeep(expected, {'a': 1.1256, 'b': 6})
        with self.assertRaises(AssertionError):
            self.assertAlmostEqualDeep(expected, [1, 2, 3])
        with self.assertRaises(AssertionError):
            self.assertAlmostEqualDeep(expected, 3456)

    def test_list(self):
        expected = ['a', 1.1256, True]
        self.assertAlmostEqualDeep(expected, expected)
        self.assertAlmostEqualDeep(expected, ['a', 1.1256, True])
        self.assertAlmostEqualDeep(expected, ('a', 1.1256, True))
        self.assertAlmostEqualDeep(expected, ['a', 1.1251, True], places=3)
        with self.assertRaises(AssertionError):
            self.assertAlmostEqualDeep(expected, ['a', 1.1251, True], places=4)
        with self.assertRaises(AssertionError):
            self.assertAlmostEqualDeep(expected, ['a', 1.1256, False], places=4)
        with self.assertRaises(AssertionError):
            self.assertAlmostEqualDeep(expected, [1, 2, 3])
        with self.assertRaises(AssertionError):
            self.assertAlmostEqualDeep(expected, 3456)

    def test_list_dict_tuple(self):
        expected = [
            {'a': True, 'b': (1.1256, 45, True)},
            {'a': False, 'b': (2.1256, 46, False)}
        ]
        self.assertAlmostEqualDeep(expected, expected)
        self.assertAlmostEqualDeep(expected, [
            {'a': True, 'b': (1.1256, 45, True)},
            {'a': False, 'b': (2.1256, 46, False)}
        ])
        self.assertAlmostEqualDeep(expected, [
            {'a': True, 'b': (1.1251, 45, True)},
            {'a': False, 'b': (2.1259, 46, False)}
        ], places=3)
        with self.assertRaises(AssertionError):
            self.assertAlmostEqualDeep(expected, [
                {'a': True, 'b': (1.1251, 45, True)},
                {'a': False, 'b': (2.1259, 46, False)}
            ], places=4)



# The MIT License (MIT)
# Copyright (c) 2023 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import functools
import gzip
import math
import sys

import numpy as np
import pandas as pd
import pyproj

from xcube.core.gridmapping import CRS_CRS84
from xcube.core.select import select_subset
from xcube.server.api import ApiError
from xcube.server.api import ApiHandler
from .api import api
from .config import DEFAULT_MAX_VOXEL_COUNT
from .context import VolumesContext
from ..datasets import PATH_PARAM_DATASET_ID
from ..datasets import PATH_PARAM_VAR_NAME


# noinspection PyPep8Naming
@api.route('/volumes/{datasetId}/{varName}')
class VolumesContextHandler(ApiHandler[VolumesContext]):
    @api.operation(operation_id='getVolume',
                   summary="Get the volume data for a variable.",
                   description=(
                           "This is an experimental feature."
                           " The volume data is returned in NRRD format."
                           " Details of the NRRD format can be found at"
                           " https://teem.sourceforge.net/nrrd/format.html"
                   ),
                   parameters=[
                       PATH_PARAM_DATASET_ID,
                       PATH_PARAM_VAR_NAME,
                       {
                           "name": "bbox",
                           "in": "query",
                           "description": 'Bounding box in degrees (WGS-84)'
                                          ' using format x1,y1,x2,y2',
                           "schema": {
                               "type": "string",
                           }
                       },
                       {
                           "name": "startDate",
                           "in": "query",
                           "description": "Start date",
                           "schema": {
                               "type": "string",
                               "format": "datetime"
                           }
                       },
                       {
                           "name": "endDate",
                           "in": "query",
                           "description": "End date",
                           "schema": {
                               "type": "string",
                               "format": "datetime"
                           }
                       },
                       {
                           "name": "encoding",
                           "in": "query",
                           "description": 'Encoding of the result',
                           "schema": {
                               "type": "string",
                               "enum": ["raw", "gz"],
                               "default": "gz",
                           }
                       },
                   ])
    async def get(self, datasetId: str, varName: str):
        bbox = self.request.get_query_arg('bbox',
                                          type=str,
                                          default='')
        if bbox:
            try:
                x1, y1, x2, y2 = map(float, [p.strip()
                                             for p in bbox.split(',')])
            except (ValueError, TypeError):
                raise ApiError.BadRequest("Invalid bbox")
            bbox = x1, y1, x2, y2
        else:
            bbox = None

        start_date = self.request.get_query_arg('startDate',
                                                type=pd.Timestamp,
                                                default=None)
        end_date = self.request.get_query_arg('endDate',
                                              type=pd.Timestamp,
                                              default=None)

        time_range = [start_date, end_date] if start_date or end_date \
            else None

        encoding = self.request.get_query_arg('encoding',
                                              type=str,
                                              default="gz")

        if encoding not in ("gz", "raw"):
            raise ApiError.BadRequest(
                'Encoding must be one of "gz" or "raw"'
            )

        ml_dataset = self.ctx.datasets_ctx.get_ml_dataset(datasetId)
        grid_mapping = ml_dataset.grid_mapping

        if bbox is not None and not grid_mapping.crs.is_geographic:
            transformer = pyproj.Transformer.from_crs(CRS_CRS84,
                                                      grid_mapping.crs)
            x1, y1, x2, y2 = bbox
            (x1, x2), (y1, y2) = transformer.transform((x1, x2), (y1, y2))
            bbox = x1, y1, x2, y2

        dataset = self.ctx.datasets_ctx.get_dataset(
            datasetId,
            expected_var_names=[varName]
        )
        var = select_subset(
            dataset,
            var_names=[varName],
            time_range=time_range,
            bbox=bbox,
            grid_mapping=ml_dataset.grid_mapping
        )[varName]

        if var.ndim != 3:
            raise ApiError.BadRequest(
                f'Variable must be 3-D, got {var.ndim}-D'
            )

        voxel_count = functools.reduce(lambda x, y: x * y, var.shape, 1)
        max_voxel_count = self.ctx.config \
            .get("VolumesAccess", {}) \
            .get("MaxVoxelCount", DEFAULT_MAX_VOXEL_COUNT)
        if voxel_count > max_voxel_count:
            raise ApiError.BadRequest(
                f'Volume too large, please select a smaller dataset subset.'
                f' Maximum is {max_voxel_count} voxels,'
                f' got {" x ".join(map(str, var.shape))} = {voxel_count}.'
            )

        # TODO (forman): allow for any dtype
        values = var.astype(dtype=np.float32).values
        if not ml_dataset.grid_mapping.is_j_axis_up:
            values = values[:, ::-1, :]
        values = np.where(np.isnan(values), 0.0, values)
        data = values.tobytes(order='C')
        if encoding == 'gz':
            data = gzip.compress(data)

        block_size = 1024 * 1024
        num_blocks = math.ceil(len(data) / block_size)

        size_z, size_y, size_x = var.shape
        # TODO (forman): find more suitable normalisation
        scale_x = scale_y = 100. / max(size_x, size_y)
        scale_z = 100. / size_z

        nrrd_header = (
            "NRRD0004\n"
            "# NRRD 4 Format\n"
            "# see http://teem.sourceforge.net/nrrd/format.html\n"
            "type: float\n"  # TODO (forman): allow for any dtype
            "dimension: 3\n"
            "sizes:"
            f" {size_x} {size_y} {size_z}\n"
            f"encoding:"
            f" {encoding}\n"
            "endian:"
            f" {sys.byteorder}\n"
            "space directions:"
            f" ({scale_x},0,0) (0,{scale_y},0) (0,0,{scale_z})\n"
            "space origin: (0,0,0)\n"
            "\n"
        )

        self.response.set_header('Content-Type', 'application/octet-stream')
        self.response.set_header('Cache-Control', 'max-age=1')
        self.response.write(bytes(nrrd_header, "utf-8"))
        import time
        for i in range(num_blocks):
            self.response.write(
                data[i * block_size: i * block_size + block_size]
            )
            time.sleep(1 / 10000)

        await self.response.finish()




# The MIT License (MIT)
# Copyright (c) 2023 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

from typing import Optional

import xarray as xr

from xcube.core.gridmapping import GridMapping
from xcube.util.assertions import assert_instance


def encode_grid_mapping(ds: xr.Dataset,
                        gm: GridMapping,
                        gm_name: Optional[str] = None,
                        force: Optional[bool] = None) -> xr.Dataset:
    """Encode the given grid mapping *gm* into a
    copy of *ds* in a CF-compliant way and return the dataset copy.
    The function removes any existing grid mappings.

    If the CRS of *gm* is geographic and the spatial dimension and coordinate
    names are "lat", "lon" and *force* is ``False``, or *force* is ``None``
    and no former grid mapping was encoded in *ds*, then nothing else is
    done and the dataset copy is returned without further action.

    Otherwise, for every spatial data variable with dims=(..., y, x),
    the function sets the attribute "grid_mapping" to *gm_name*.
    The grid mapping CRS is encoded in a new 0-D variable named *gm_name*.

    :param ds: The dataset.
    :param gm: The dataset's grid mapping.
    :param gm_name: Name for the grid mapping variable.
        Defaults to "crs".
    :param force: Whether to force encoding of grid mapping even
        if CRS is geographic and spatial dimension names are "lon", "lat".
        Optional value, if not provided, *force* will be assumed ``True``
        if a former grid mapping was encoded in *ds*.
    :return: A copy of *ds* with *gm* encoded into it.
    """
    assert_instance(ds, xr.Dataset, "ds")
    assert_instance(gm, GridMapping, "gm")
    if gm_name is not None:
        assert_instance(gm_name, str, "gm_name")

    ds_copy = ds.copy()

    x_dim_name, y_dim_name = gm.xy_dim_names
    spatial_vars = [(var_name, var)
                    for var_name, var in ds.data_vars.items()
                    if (var.ndim >= 2
                        and var.dims[-1] == x_dim_name
                        and var.dims[-2] == y_dim_name)]

    old_gm_names = set(old_gm_name
                       for old_gm_name in (
                           var.attrs.get("grid_mapping")
                           for var_name, var in spatial_vars
                       )
                       if old_gm_name and old_gm_name in ds_copy)
    if old_gm_names:
        force = True if force is None else force
        gm_name = gm_name or next(iter(old_gm_names))
        ds_copy = ds_copy.drop_vars(old_gm_names)

    is_geographic = (gm.xy_var_names == gm.xy_dim_names
                     and gm.xy_dim_names == ("lon", "lat")
                     and gm.crs.is_geographic)

    if force or not is_geographic:
        gm_name = gm_name or "crs"
        for var_name, var in spatial_vars:
            ds_copy[var_name] = var.assign_attrs(grid_mapping=gm_name)
        ds_copy[gm_name] = xr.DataArray(0, attrs=gm.crs.to_cf())

    return ds_copy


def maybe_encode_grid_mapping(encode_cf: bool,
                              ds: xr.Dataset,
                              gm: GridMapping,
                              gm_name: Optional[str]) -> xr.Dataset:
    """Internal helper."""
    if encode_cf:
        return encode_grid_mapping(
            ds, gm,
            gm_name=gm_name,
            force=True if gm_name else None
        )
    return ds



# The MIT License (MIT)
# Copyright (c) 2019 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import functools
import operator
from typing import Sequence, Optional, Tuple

import click

from xcube.cli.common import (parse_cli_sequence,
                              assert_positive_int_item,
                              cli_option_quiet,
                              cli_option_verbosity, configure_cli_output)
from xcube.constants import FORMAT_NAME_ZARR, FORMAT_NAME_NETCDF4, \
    FORMAT_NAME_MEM
from xcube.constants import LOG

OUTPUT_FORMAT_NAMES = [FORMAT_NAME_ZARR, FORMAT_NAME_NETCDF4, FORMAT_NAME_MEM]

DEFAULT_OUTPUT_PATH = 'out.zarr'
DEFAULT_XY_NAMES = 'lon,lat'
DEFAULT_DELTA = 1e-5
DEFAULT_DRY_RUN = False
DEFAULT_CRS = 'EPSG:4326'


# noinspection PyShadowingBuiltins
@click.command(name='rectify', hidden=True)
@click.argument('dataset', metavar='INPUT')
@click.option('--xy-vars', 'xy_var_names',
              help=f'Comma-separated names of variables providing x any y coordinates. '
                   f'If omitted, names will be guessed from available coordinate variables in INPUT, e.g. "lon,lat".')
@click.option('--var', '-v', 'var_names', multiple=True, metavar='VARIABLES',
              help="Comma-separated list of names of variables to be included or multiple options may be given. "
                   "If omitted, all variables in INPUT will be reprojected.")
@click.option('--output', '-o', 'output_path', metavar='OUTPUT',
              default=DEFAULT_OUTPUT_PATH,
              help=f"Output path. Defaults to {DEFAULT_OUTPUT_PATH!r}.")
@click.option('--format', '-f', 'output_format', metavar='FORMAT',
              type=click.Choice(OUTPUT_FORMAT_NAMES),
              help="Output format. "
                   "If omitted, format will be guessed from OUTPUT.")
@click.option('--size', '-s', 'output_size', metavar='SIZE',
              help='Output size in pixels using format "WIDTH,HEIGHT", e.g. "2048,1024". '
                   'If omitted, a size will be computed so spatial resolution of INPUT is preserved.')
@click.option('--tile_size', '-t', 'output_tile_size', metavar='TILE_SIZE',
              help='Output tile size in pixels using format "WIDTH,HEIGHT", e.g. "512,512". '
                   'If given, the output will be chunked w.r.t. TILE_SIZE. Otherwise the output will NOT be chunked.')
@click.option('--point', '-p', 'output_point', metavar='POINT',
              help='Output spatial coordinates of the point referring to pixel col=0.5,row=0.5 '
                   'using format "LON,LAT" or "X,Y", e.g. "1.2,53.5". '
                   'If omitted, the default reference point will the INPUT\'s minimum spatial coordinates.')
@click.option('--res', '-r', 'output_res', metavar='RES', type=float,
              help='Output spatial resolution in decimal degrees. '
                   'If omitted, the default resolution will be close to the spatial resolution of INPUT.')
@click.option('--crs', '-r', 'output_crs', metavar='CRS',
              help='Output spatial coordinate reference system (CRS). '
                   f'If omitted, the default CRS will be "{DEFAULT_CRS}".')
@click.option('--delta', '-d', type=float, default=DEFAULT_DELTA,
              help='Relative maximum delta for detection whether a '
                   'target pixel center is within a source pixel\'s boundary.')
@cli_option_quiet
@cli_option_verbosity
@click.option('--dry-run', default=DEFAULT_DRY_RUN, is_flag=True,
              help='Just read and process INPUT, but don\'t produce any outputs.')
def rectify(dataset: str,
            xy_var_names: str = None,
            var_names: str = None,
            output_path: str = None,
            output_format: str = None,
            output_size: str = None,
            output_tile_size: str = None,
            output_point: str = None,
            output_res: float = None,
            output_crs: str = None,
            delta: float = DEFAULT_DELTA,
            quiet: bool = False,
            verbosity: int = 0,
            dry_run: bool = DEFAULT_DRY_RUN):
    """
    Rectify a dataset to WGS-84 using its per-pixel geo-locations.
    """
    configure_cli_output(quiet=quiet, verbosity=verbosity)

    input_path = dataset

    xy_var_names = parse_cli_sequence(xy_var_names,
                                      metavar='VARIABLES', num_items=2,
                                      item_plural_name='names')
    var_name_lists = [parse_cli_sequence(var_name_specifier,
                                         metavar='VARIABLES',
                                         item_plural_name='names')
                      for var_name_specifier in var_names]
    var_name_flat_list = functools.reduce(operator.iconcat, var_name_lists,
                                          [])

    output_size = parse_cli_sequence(output_size,
                                     metavar='SIZE', num_items=2, item_plural_name='sizes',
                                     item_parser=int, item_validator=assert_positive_int_item)
    output_tile_size = parse_cli_sequence(output_tile_size,
                                          metavar='TILE_SIZE', num_items=2, item_plural_name='tile sizes',
                                          item_parser=int, item_validator=assert_positive_int_item)
    output_point = parse_cli_sequence(output_point,
                                      metavar='POINT', num_items=2, item_plural_name='coordinates',
                                      item_parser=float)

    # noinspection PyBroadException
    _rectify(input_path,
             xy_var_names,
             None if len(var_name_flat_list) == 0 else var_name_flat_list,
             output_path,
             output_format,
             output_size,
             output_tile_size,
             output_point,
             output_res,
             output_crs,
             delta,
             dry_run=dry_run,
             monitor=LOG.info)

    return 0


def _rectify(input_path: str,
             xy_names: Optional[Tuple[str, str]],
             var_names: Optional[Sequence[str]],
             output_path: str,
             output_format: Optional[str],
             output_size: Optional[Tuple[int, int]],
             output_tile_size: Optional[Tuple[int, int]],
             output_point: Optional[Tuple[float, float]],
             output_res: Optional[float],
             output_crs: Optional[str],
             delta: float,
             dry_run: bool,
             monitor):
    import pyproj.crs

    from xcube.core.dsio import guess_dataset_format
    from xcube.core.dsio import open_dataset
    from xcube.core.dsio import write_dataset
    from xcube.core.gridmapping import GridMapping
    from xcube.core.resampling import rectify_dataset
    from xcube.core.sentinel3 import is_sentinel3_product
    from xcube.core.sentinel3 import open_sentinel3_product

    if not output_format:
        output_format = guess_dataset_format(output_path)

    output_gm = None
    output_gm_given = (output_size is not None,
                       output_point is not None,
                       output_res is not None,
                       output_crs is not None)
    if all(output_gm_given):
        output_gm = GridMapping.regular(size=output_size,
                                        xy_min=output_point,
                                        xy_res=output_res,
                                        crs=pyproj.crs.CRS.from_user_input(output_crs))
    elif any(output_gm_given):
        raise click.ClickException('SIZE, POINT, RES, and CRS must all be given or none of them.')

    monitor(f'Opening dataset from {input_path!r}...')

    if is_sentinel3_product(input_path):
        src_ds = open_sentinel3_product(input_path)
    else:
        src_ds = open_dataset(input_path)

    monitor('Rectifying...')
    rectified_ds = rectify_dataset(src_ds,
                                   xy_var_names=xy_names,
                                   var_names=var_names,
                                   target_gm=output_gm,
                                   tile_size=output_tile_size,
                                   uv_delta=delta)

    if rectified_ds is None:
        monitor(f'Dataset {input_path} does not seem to have an intersection with bounding box')
        return

    monitor(f'Writing rectified dataset to {output_path!r}...')
    if not dry_run:
        write_dataset(rectified_ds, output_path, output_format)

    monitor(f'Done.')



# The MIT License (MIT)
# Copyright (c) 2022 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

from xcube.server.api import ApiError
from xcube.server.api import ApiHandler
from .api import api
from .context import WmtsContext
from .controllers import WMTS_CRS84_TMS_ID
from .controllers import WMTS_TILE_FORMAT
from .controllers import WMTS_VERSION
from .controllers import WMTS_WEB_MERCATOR_TMS_ID
from .controllers import get_crs_name_from_tms_id
from .controllers import get_wmts_capabilities_xml
from ...datasets import PATH_PARAM_DATASET_ID
from ...datasets import PATH_PARAM_VAR_NAME
from ...datasets import QUERY_PARAM_CBAR
from ...datasets import QUERY_PARAM_CRS
from ...datasets import QUERY_PARAM_VMAX
from ...datasets import QUERY_PARAM_VMIN
from ...tiles import PATH_PARAM_X
from ...tiles import PATH_PARAM_Y
from ...tiles import PATH_PARAM_Z
from ...tiles import QUERY_PARAM_RETINA
from ...tiles import QUERY_PARAM_TIME
from ...tiles.controllers import compute_ml_dataset_tile

_VALID_WMTS_TMS_IDS = (WMTS_CRS84_TMS_ID, WMTS_WEB_MERCATOR_TMS_ID)

PATH_PARAM_TMS_ID = {
    "name": "tmsId",
    "in": "query",
    "description": "Tile matrix set identifier",
    "schema": {
        "type": "string",
        "enum": [WMTS_CRS84_TMS_ID, WMTS_WEB_MERCATOR_TMS_ID],
    }
}

TILE_PARAMETERS = [
    PATH_PARAM_DATASET_ID,
    PATH_PARAM_VAR_NAME,
    PATH_PARAM_Z,
    PATH_PARAM_Y,
    PATH_PARAM_X,
    QUERY_PARAM_CRS,
    QUERY_PARAM_VMIN,
    QUERY_PARAM_VMAX,
    QUERY_PARAM_CBAR,
    QUERY_PARAM_TIME,
    QUERY_PARAM_RETINA,
]

TMS_TILE_PARAMETERS = [
    PATH_PARAM_DATASET_ID,
    PATH_PARAM_VAR_NAME,
    PATH_PARAM_Z,
    PATH_PARAM_Y,
    PATH_PARAM_X,
    QUERY_PARAM_CRS,
    QUERY_PARAM_VMIN,
    QUERY_PARAM_VMAX,
    QUERY_PARAM_CBAR,
    QUERY_PARAM_TIME,
    QUERY_PARAM_RETINA,
]


@api.route('/wmts/1.0.0/WMTSCapabilities.xml')
class WmtsCapabilitiesXmlHandler(ApiHandler[WmtsContext]):
    @api.operation(operationId='getWmtsCapabilities',
                   summary='Gets the WMTS capabilities as XML document')
    async def get(self):
        self.request.make_query_lower_case()
        capabilities = await self.ctx.run_in_executor(
            None,
            get_wmts_capabilities_xml,
            self.ctx,
            self.request.reverse_base_url,
            WMTS_CRS84_TMS_ID
        )
        self.response.set_header('Content-Type', 'application/xml')
        await self.response.finish(capabilities)


@api.route('/wmts/1.0.0/{tmsId}/WMTSCapabilities.xml')
class WmtsCapabilitiesXmlForTmsHandler(ApiHandler[WmtsContext]):
    # noinspection PyPep8Naming
    @api.operation(operationId='getWmtsTmsCapabilities',
                   summary='Gets the WMTS capabilities'
                           ' for tile matrix set as XML document',
                   parameters=[PATH_PARAM_TMS_ID])
    async def get(self, tmsId):
        self.request.make_query_lower_case()
        _assert_valid_tms_id(tmsId)
        capabilities = await self.ctx.run_in_executor(
            None,
            get_wmts_capabilities_xml,
            self.ctx,
            self.request.reverse_base_url,
            tmsId
        )
        self.response.set_header('Content-Type', 'application/xml')
        await self.response.finish(capabilities)


@api.route('/wmts/1.0.0/tile/{datasetId}/{varName}/{z}/{y}/{x}.png')
class WmtsImageTileHandler(ApiHandler[WmtsContext]):
    # noinspection PyPep8Naming
    @api.operation(operationId='getWmtsImageTile',
                   summary='Gets a WMTS image tile in PNG format.',
                   parameters=TILE_PARAMETERS)
    async def get(self,
                  datasetId: str,
                  varName: str,
                  z: str, y: str, x: str):
        self.request.make_query_lower_case()
        tms_id = self.request.get_query_arg(
            'tilematrixset', default=WMTS_CRS84_TMS_ID
        )
        _assert_valid_tms_id(tms_id)
        crs_name = get_crs_name_from_tms_id(tms_id)
        tile = await self.ctx.run_in_executor(
            None,
            compute_ml_dataset_tile,
            self.ctx.tiles_ctx,
            datasetId,
            varName,
            crs_name,
            x, y, z,
            _query_to_dict(self.request)
        )
        self.response.set_header('Content-Type', 'image/png')
        await self.response.finish(tile)


@api.route('/wmts/1.0.0/tile/{datasetId}/{varName}/{tmsId}/{z}/{y}/{x}.png')
class WmtsImageTileForTmsHandler(ApiHandler[WmtsContext]):
    # noinspection PyPep8Naming
    @api.operation(operationId='getWmtsTmsImageTile',
                   summary='Gets a WMTS image tile'
                           ' for given tile matrix set in PNG format',
                   parameters=TMS_TILE_PARAMETERS)
    async def get(self,
                  datasetId: str,
                  varName: str,
                  tmsId: str,
                  z: str, y: str, x: str):
        self.request.make_query_lower_case()
        _assert_valid_tms_id(tmsId)
        crs_name = get_crs_name_from_tms_id(tmsId)
        tile = await self.ctx.run_in_executor(
            None,
            compute_ml_dataset_tile,
            self.ctx.tiles_ctx,
            datasetId,
            varName,
            crs_name,
            x, y, z,
            _query_to_dict(self.request)
        )
        self.response.set_header('Content-Type', 'image/png')
        await self.response.finish(tile)


@api.route('/wmts/kvp')
class WmtsKvpHandler(ApiHandler[WmtsContext]):
    @api.operation(operationId='invokeWmtsMethodFromKvp',
                   summary='Invokes the WMTS by key-value pairs',
                   parameters=TILE_PARAMETERS)
    async def get(self):
        self.request.make_query_lower_case()
        service = self.request.get_query_arg('service')
        if service != "WMTS":
            raise ApiError.BadRequest(
                'value for "service" parameter must be "WMTS"'
            )
        request = self.request.get_query_arg('request')
        if request == "GetCapabilities":
            wmts_version = self.request.get_query_arg(
                "version", default=WMTS_VERSION
            )
            if wmts_version != WMTS_VERSION:
                raise ApiError.BadRequest(
                    f'value for "version" parameter must be "{WMTS_VERSION}"'
                )
            tms_id = self.request.get_query_arg(
                "tilematrixset", default=WMTS_CRS84_TMS_ID
            )
            _assert_valid_tms_id(tms_id)
            capabilities = await self.ctx.run_in_executor(
                None,
                get_wmts_capabilities_xml,
                self.ctx,
                self.request.reverse_base_url,
                tms_id
            )
            self.response.set_header("Content-Type", "application/xml")
            await self.response.finish(capabilities)

        elif request == "GetTile":
            wmts_version = self.request.get_query_arg("version",
                                                      default=WMTS_VERSION)
            if wmts_version != WMTS_VERSION:
                raise ApiError.BadRequest(
                    f'value for "version" parameter must be "{WMTS_VERSION}"'
                )
            layer = self.request.get_query_arg("layer")
            try:
                ds_id, var_name = layer.split(".")
            except ValueError as e:
                raise ApiError.BadRequest(
                    'value for "layer" parameter must be'
                    ' "<dataset>.<variable>"'
                ) from e
            # For time being, we ignore "style"
            # style = self.request.get_query_arg("style"
            mime_type = self.request.get_query_arg(
                "format", default=WMTS_TILE_FORMAT
            ).lower()
            if mime_type not in (WMTS_TILE_FORMAT, "png"):
                raise ApiError.BadRequest(
                    f'value for "format" parameter'
                    f' must be "{WMTS_TILE_FORMAT}"'
                )
            tms_id = self.request.get_query_arg(
                'tilematrixset', default=WMTS_CRS84_TMS_ID
            )
            _assert_valid_tms_id(tms_id)
            crs_name = get_crs_name_from_tms_id(tms_id)
            x = self.request.get_query_arg("tilecol", type=int)
            y = self.request.get_query_arg("tilerow", type=int)
            z = self.request.get_query_arg("tilematrix", type=int)
            tile = await self.ctx.run_in_executor(
                None,
                compute_ml_dataset_tile,
                self.ctx.tiles_ctx,
                ds_id,
                var_name,
                crs_name,
                x, y, z,
                _query_to_dict(self.request)
            )
            self.response.set_header("Content-Type", "image/png")
            await self.response.finish(tile)
        elif request == "GetFeatureInfo":
            raise ApiError.BadRequest(
                'request type "GetFeatureInfo" not yet implemented'
            )
        else:
            raise ApiError.BadRequest(
                f'invalid request type "{request}"'
            )


def _assert_valid_tms_id(tms_id: str):
    if tms_id not in _VALID_WMTS_TMS_IDS:
        raise ApiError.BadRequest(
            f'value for "tilematrixset" parameter'
            f' must be one of {_VALID_WMTS_TMS_IDS!r}'
        )


def _query_to_dict(request):
    return {k: v[0] for k, v in request.query.items()}



# The MIT License (MIT)
# Copyright (c) 2022 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import threading
from typing import Any, Dict, List, Optional, Mapping, Iterator, Union, Tuple

import fsspec

from xcube.constants import LOG
from xcube.server.api import ApiContext
from xcube.server.api import ApiError
from xcube.server.api import Context
from xcube.server.config import get_base_dir
from xcube.server.config import resolve_config_path
from xcube.util.perf import measure_time_cm
from xcube.version import version
from xcube.webapi.auth import AuthContext


class ResourcesContext(ApiContext):

    def __init__(self, server_ctx: Context):
        super().__init__(server_ctx)
        # noinspection PyTypeChecker
        self._auth_ctx: AuthContext = server_ctx.get_api_ctx("auth")
        assert isinstance(self._auth_ctx, AuthContext)
        self._base_dir = get_base_dir(self.config)
        self._prefix = normalize_prefix(self.config.get("prefix", ""))
        self._trace_perf = self.config.get("trace_perf", False)
        self._rlock = threading.RLock()

    @property
    def auth_ctx(self) -> AuthContext:
        return self._auth_ctx

    @property
    def base_dir(self) -> str:
        return self._base_dir

    @property
    def rlock(self) -> threading.RLock:
        return self._rlock

    @property
    def trace_perf(self) -> bool:
        return self._trace_perf

    @property
    def measure_time(self):
        return measure_time_cm(disabled=not self.trace_perf, logger=LOG)

    @property
    def can_authenticate(self) -> bool:
        """
        Test whether the user can authenticate.
        Even if authentication service is configured, user authentication
        may still be optional. In this case the server will publish
        the resources configured to be free for everyone.
        """
        return self._auth_ctx.can_authenticate

    @property
    def must_authenticate(self) -> bool:
        """
        Test whether the user must authenticate.
        """
        return self._auth_ctx.must_authenticate

    @property
    def access_control(self) -> Dict[str, Any]:
        return self.config.get('AccessControl', {})

    @property
    def required_scopes(self) -> List[str]:
        return self.access_control.get('RequiredScopes', [])

    def get_service_url(self, base_url: Optional[str], *path: str):
        base_url = base_url or ''
        # noinspection PyTypeChecker
        path_comp = '/'.join(path)
        if self._prefix:
            return base_url + self._prefix + '/' + path_comp
        else:
            return base_url + '/' + path_comp

    def get_config_path(self,
                        config: Mapping[str, Any],
                        config_name: str,
                        path_entry_name: str = 'Path') -> str:
        path = config.get(path_entry_name)
        if not path:
            raise ApiError.InvalidServerConfig(
                f"Missing entry {path_entry_name!r} in {config_name}"
            )
        return self.resolve_config_path(path)

    def resolve_config_path(self, path: str) -> str:
        return resolve_config_path(self.config, path)

    def load_file(self,
                  path: str,
                  mode: str = "rb",
                  encoding: str = "utf-8") -> Union[str, bytes]:
        with self.open_file(path, mode=mode, encoding=encoding) as fp:
            return fp.read()

    def open_file(self,
                  path: str,
                  mode: str = "rb",
                  encoding: str = "utf-8") -> fsspec.core.OpenFile:
        path = self.resolve_config_path(path)
        return fsspec.open(path, mode=mode, encoding=encoding)

    def eval_config_value(self, value: Any) -> Any:
        """Evaluate expressions in the given *value* if it is
        a string. If *value* is a dict or list, then the method
        is called recursively and applied to the values of the
        dict or the items of the list. In this case a dict or list
        will be returned.
        Expressions are any string between non-quoted "${" and "}".
        An iterator is returned comprising non-expression
        parts as well as parts evaluated in the context of this
        context object.
        """
        if isinstance(value, dict):
            return {k: self.eval_config_value(v)
                    for k, v in value.items()}
        if isinstance(value, list):
            return [self.eval_config_value(v)
                    for v in value]
        if isinstance(value, str):
            return self._eval_config_str_value(value)
        return value

    def _eval_config_str_value(self, value: str) -> Any:
        values = list(self._eval_value(value))
        if len(values) > 1:
            return "".join([str(v) for v in values])
        elif len(values) == 1:
            return values[0]
        return value

    def _eval_value(self, value: str) -> Iterator[Any]:
        """Evaluate expressions in a given string *value*.
        Expressions are any string between non-quoted "${" and "}".
        An iterator is returned comprising non-expression
        parts as well as parts evaluated in the context of this
        context object.
        """
        for token in self._tokenize_value(value):
            if isinstance(token, tuple):
                expression, = token
                yield eval(expression, None, self.new_eval_env())
            else:
                yield token

    def new_eval_env(self) -> Dict[str, Any]:
        return dict(
            ctx=self,
            base_dir=self.base_dir,
            load_file=self.load_file,
            open_file=self.open_file,
            resolve_config_path=self.resolve_config_path
        )

    @classmethod
    def _tokenize_value(cls, value: str) -> Iterator[Union[str, Tuple[str]]]:
        """Tokenize a string value. Tokens are either strings or 1-tuples
        that contain a string-expression to be evaluated. String-expressions
        are any strings between non-quoted "${" and "}".
        """
        n = len(value)
        i0 = 0
        sqs = False
        dqs = False
        es = False
        for i in range(n):
            c = value[i]
            if c == "\\":
                continue
            elif c == "'" and not dqs:
                sqs = not sqs
            elif c == '"' and not sqs:
                dqs = not dqs
            elif c == '$' and not (sqs or dqs or es):
                if i < n - 1 and value[i + 1] == '{':
                    if i0 < i:
                        yield value[i0:i]  # Return a str.
                    es = True
                    i0 = i + 2
                    i += 1
            elif c == '}' and es:
                es = False
                yield value[i0:i],  # Yes, return a 1-tuple.
                i0 = i + 1
        if i0 == 0:
            yield value
        elif i0 < n:
            yield value[i0:n]  # Return a str.


def normalize_prefix(prefix: Optional[str]) -> str:
    if not prefix:
        return ''

    prefix = prefix.replace('${name}', 'xcube')
    prefix = prefix.replace('${version}', version)
    prefix = prefix.replace('//', '/').replace('//', '/')

    if prefix == '/':
        return ''

    if not prefix.startswith('/'):
        prefix = '/' + prefix

    if prefix.endswith('/'):
        prefix = prefix[0:-1]

    return prefix



# The MIT License (MIT)
# Copyright (c) 2021 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import numpy as np
import xarray as xr


def is_empty_cube(cube: xr.Dataset) -> bool:
    return len(cube.data_vars) == 0


def strip_cube(cube: xr.Dataset) -> xr.Dataset:
    drop_vars = [k for k, v in cube.data_vars.items()
                 if len(v.shape) < 3
                 or np.product(v.shape) == 0
                 or v.shape[-2] < 2
                 or v.shape[-1] < 2]
    if drop_vars:
        return cube.drop_vars(drop_vars)
    return cube



# The MIT License (MIT)
# Copyright (c) 2019 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import functools
import logging
import time
import warnings
from contextlib import AbstractContextManager
from typing import Optional

from xcube.constants import LOG


def measure_time_cm(logger=None, disabled=False):
    """
    Get a context manager for measuring execution time of code blocks and logging the result.

    Measure duration and log output:::

        measure_time = measure_time_cm()
        with measure_time("heavy computation"):
            do_heavy_computation()

    or just measure duration:::

        with measure_time() as cm:
            do_heavy_computation()
        print("heavy computation took %2.f seconds" % cm.duration)

    :param logger: The logger to be used. May be a string or logger object. Defaults to "xcube".
    :param disabled: If True, efficiently disables timing and logging.
    :return: a context manager callable
    """
    if disabled:
        return _do_not_measure_time_cm
    else:
        return functools.partial(measure_time, logger=logger)


class measure_time(AbstractContextManager):
    def __init__(self,
                 *args,
                 tag: Optional[str] = None,
                 logger: Optional[logging.Logger] = None,
                 **kwargs):
        self.message = tag
        self.args = args
        self.kwargs = kwargs
        if tag is not None:
            warnings.warn('The keyword "tag" has been deprecated,'
                          ' use first argument "message" instead',
                          DeprecationWarning)
        elif args:
            self.message = args[0]
            self.args = args[1:]
        else:
            warnings.warn('Calling measure_time() without "message"'
                          ' argument is deprecated.',
                          DeprecationWarning)
            self.message = None
            self.args = None
        if isinstance(logger, str):
            self.logger = logging.getLogger(logger)
        elif logger is None:
            self.logger = LOG
        else:
            self.logger = logger
        self.start_time = None
        self.duration = None

    def __enter__(self):
        self.start_time = time.perf_counter()
        return self

    def __exit__(self, *exc):
        self.duration = time.perf_counter() - self.start_time
        if self.message:
            self.logger.info(
                self.message + ": took %.2fms",
                *self.args,
                self.duration * 1000,
                **self.kwargs
            )


class _do_not_measure_time_cm(AbstractContextManager):

    # noinspection PyUnusedLocal
    def __init__(self, *args, **kwargs):
        self.duration = None
        self.logger = None
        self.message = None

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        pass



# The MIT License (MIT)
# Copyright (c) 2022 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import hashlib
from typing import Optional

from xcube.server.api import ApiHandler, ApiError
from .api import api
from .context import S3Context
from .listbucket import _LAST_MODIFIED_DUMMY
from .listbucket import dict_to_xml
from .listbucket import list_bucket_result_to_xml
from .listbucket import list_s3_bucket_v1
from .listbucket import list_s3_bucket_v2

BUCKET_PARAMETER = {
    "name": "bucket",
    "in": "path",
    "description": 'The bucket name.'
                   ' The bucket "datasets" provides Zarr datasets,'
                   ' the bucket "pyramids" provides them as'
                   ' multi-resolution image pyramids.',
    "schema": {
        "type": "string",
        "enum": ["datasets", "pyramids"],
    }
}

OPTIONAL_STRING_SCHEMA =   {
            "type": ["string", "null"],
            "default": None,
        }

LIST_BUCKET_PARAMETERS = [
    BUCKET_PARAMETER,
    {
        "name": "list-type",
        "in": "query",
        "description": "The operation version (1 or 2)",
        "schema": {
            "type": ["integer", "null"],
            "enum": [None, 1, 2],
            "default": None,
        }
    },
    {
        "name": "prefix",
        "in": "query",
        "description": "Key prefix",
        "schema": OPTIONAL_STRING_SCHEMA
    },
    {
        "name": "delimiter",
        "in": "query",
        "description": "Key delimiter",
        "schema": OPTIONAL_STRING_SCHEMA
    },
    {
        "name": "max-keys",
        "in": "query",
        "description": "Maximum number of keys to return",
        "schema": {
            "type": "integer",
            "default": 1000,
        }
    },
    {
        "name": "start-after",
        "in": "query",
        "description": "Start after given key. For list-type=2 only.",
        "schema": OPTIONAL_STRING_SCHEMA
    },
    {
        "name": "continuation-token",
        "in": "query",
        "description": "Continuation token for paging. For list-type=2 only.",
        "schema": OPTIONAL_STRING_SCHEMA
    },
    {
        "name": "marker",
        "in": "query",
        "description": "Continuation token for paging. For list-type=1 only.",
        "schema": OPTIONAL_STRING_SCHEMA
    },
]

GET_OBJECT_PARAMETERS = [
    BUCKET_PARAMETER,
    {
        "name": "key",
        "in": "path",
        "description": "The object's key.",
        "schema": {
            "type": "string",
        }
    },
]


@api.route('/s3/{bucket}')
class ListS3BucketHandler(ApiHandler[S3Context]):
    @api.operation(operationId='listS3Bucket',
                   summary='List bucket contents.',
                   parameters=LIST_BUCKET_PARAMETERS)
    async def get(self, bucket: str):
        try:
            object_storage = self.ctx.get_bucket(bucket)
        except KeyError:
            return await self._bucket_not_found(bucket)

        list_s3_bucket_params = dict(
            prefix=self.request.get_query_arg(
                'prefix', default=None
            ),
            delimiter=self.request.get_query_arg(
                'delimiter', default=None
            ),
            max_keys=int(self.request.get_query_arg(
                'max-keys', default='1000')
            )
        )

        list_type = self.request.get_query_arg('list-type', default=None)
        if list_type == '2':  # Most frequently used
            list_s3_bucket = list_s3_bucket_v2
            list_s3_bucket_params.update(
                start_after=self.request.get_query_arg(
                    'start-after', default=None
                ),
                continuation_token=self.request.get_query_arg(
                    'continuation-token', default=None
                )
            )
        elif list_type in (None, '1'):  # Old clients
            list_s3_bucket = list_s3_bucket_v1
            list_s3_bucket_params.update(
                marker=self.request.get_query_arg(
                    'marker', default=None
                )
            )
        else:
            raise ApiError.BadRequest(
                f'Unknown bucket list type {list_type!r}'
            )

        list_bucket_result = list_s3_bucket(object_storage,
                                            name=bucket,
                                            **list_s3_bucket_params)

        xml = list_bucket_result_to_xml(list_bucket_result)
        self.response.set_header('Content-Type', 'application/xml')
        await self.response.finish(xml)


@api.route('/s3/{bucket}/{*key}')
class GetS3BucketObjectHandler(ApiHandler[S3Context]):
    @api.operation(operationId='getS3ObjectMetadata',
                   summary='Get object metadata for given key.',
                   parameters=GET_OBJECT_PARAMETERS)
    async def head(self,
                   bucket: str,
                   key: Optional[str]):
        try:
            object_storage = self.ctx.get_bucket(bucket)
        except KeyError:
            return await self._bucket_not_found(bucket)

        try:
            value = object_storage[key]
        except KeyError:
            return await self._key_not_found(key)

        e_tag = hashlib.md5(value).hexdigest()
        content_length = len(value)

        self.response.set_header('ETag', f'"{e_tag}"')
        self.response.set_header('Last-Modified', _LAST_MODIFIED_DUMMY)
        self.response.set_header('Content-Length', str(content_length))
        await self.response.finish()

    # noinspection PyPep8Naming
    @api.operation(operationId='getS3ObjectData',
                   summary='Get object for given key.',
                   parameters=GET_OBJECT_PARAMETERS)
    async def get(self,
                  bucket: str,
                  key: Optional[str]):
        try:
            object_storage = self.ctx.get_bucket(bucket)
        except KeyError:
            return await self._bucket_not_found(bucket)

        try:
            value = object_storage[key]
        except KeyError:
            return await self._key_not_found(key)

        e_tag = hashlib.md5(value).hexdigest()
        content_length = len(value)

        self.response.set_header('ETag', f'"{e_tag}"')
        self.response.set_header('Last-Modified', _LAST_MODIFIED_DUMMY)
        self.response.set_header('Content-Length', str(content_length))
        self.response.set_header('Content-Type', 'binary/octet-stream')
        await self.response.finish(value)

    def _key_not_found(self, key: str):
        return self._not_found("NoSuchKey",
                               "The specified key does not exist.",
                               Key=key)

    def _bucket_not_found(self, bucket_name: str):
        return self._not_found("NoSuchBucket",
                               "The specified bucket does not exist.",
                               BucketName=bucket_name)

    def _not_found(self, code: str, message: str, **kwargs):
        self.response.set_header('Content-Type', 'application/xml')
        self.response.set_status(404)
        return self.response.finish(dict_to_xml(
            root_element_name="Error",
            content_dict={
                "Code": code,
                "Message": message,
                **kwargs
            }
        ))



# The MIT License (MIT)
# Copyright (c) 2019 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import datetime
from typing import Any
from typing import Dict

import xarray as xr

from xcube.constants import FORMAT_NAME_NETCDF4
from xcube.constants import FORMAT_NAME_ZARR
from xcube.core.gridmapping import GridMapping
from xcube.util.config import NameDictPairList

_TIME_ATTRS_DATA = ('time', 'time_bnds',
                    ('time_coverage_start', 'time_coverage_end'),
                    str)


def update_dataset_attrs(dataset: xr.Dataset,
                         global_attrs: Dict[str, Any] = None,
                         update_existing: bool = False,
                         in_place: bool = False) -> xr.Dataset:
    """
    Update spatio-temporal CF/THREDDS attributes given *dataset* according
    to spatio-temporal coordinate variables time, lat, and lon.

    :param dataset: The dataset.
    :param global_attrs: Optional global attributes.
    :param update_existing: If ``True``, any existing attributes will be updated.
    :param in_place: If ``True``, *dataset* will be modified in place and returned.
    :return: A new dataset, if *in_place* if ``False`` (default), else the passed and modified *dataset*.
    """
    if not in_place:
        dataset = dataset.copy()

    if global_attrs:
        dataset.attrs.update(global_attrs)

    dataset = update_dataset_spatial_attrs(dataset,
                                           update_existing=update_existing,
                                           in_place=in_place)
    dataset = update_dataset_temporal_attrs(dataset,
                                            update_existing=update_existing,
                                            in_place=in_place)
    return dataset


def update_dataset_spatial_attrs(dataset: xr.Dataset,
                                 update_existing: bool = False,
                                 in_place: bool = False) -> xr.Dataset:
    """
    Update spatial CF/THREDDS attributes of given *dataset*.

    :param dataset: The dataset.
    :param update_existing: If ``True``, any existing attributes will be updated.
    :param in_place: If ``True``, *dataset* will be modified in place and returned.
    :return: A new dataset, if *in_place* if ``False`` (default), else the passed and modified *dataset*.
    """
    if not in_place:
        dataset = dataset.copy()
    gm = GridMapping.from_dataset(dataset)
    gs_attrs = {
        'geospatial_lon_min',
        'geospatial_lon_max',
        'geospatial_lat_min',
        'geospatial_lat_max',
    }
    if update_existing or not gs_attrs.issubset(dataset.attrs):
        # Update dataset with newly retrieved attributes
        dataset.attrs.update(gm.to_dataset_attrs())
        dataset.attrs['date_modified'] = datetime.datetime.now().isoformat()
    return dataset


def update_dataset_temporal_attrs(dataset: xr.Dataset,
                                  update_existing: bool = False,
                                  in_place: bool = False) -> xr.Dataset:
    """
    Update temporal CF/THREDDS attributes of given *dataset*.

    :param dataset: The dataset.
    :param update_existing: If ``True``, any existing attributes will be updated.
    :param in_place: If ``True``, *dataset* will be modified in place and returned.
    :return: A new dataset, if *in_place* is ``False`` (default), else the passed and modified *dataset*.
    """
    coord_data = [_TIME_ATTRS_DATA]
    if not in_place:
        dataset = dataset.copy()

    for coord_name, coord_bnds_name, coord_attr_names, cast in coord_data:
        coord_min_attr_name, coord_max_attr_name = coord_attr_names
        if update_existing or \
                coord_min_attr_name not in dataset.attrs or \
                coord_max_attr_name not in dataset.attrs:
            coord = None
            coord_bnds = None
            coord_res = None
            if coord_name in dataset:
                coord = dataset[coord_name]
                coord_bnds_name = coord.attrs.get('bounds', coord_bnds_name)
            if coord_bnds_name in dataset:
                coord_bnds = dataset[coord_bnds_name]
            if coord_bnds is not None \
                    and coord_bnds.ndim == 2 \
                    and coord_bnds.shape[0] > 0 \
                    and coord_bnds.shape[1] == 2:
                coord_v1 = coord_bnds[0][0]
                coord_v2 = coord_bnds[-1][1]
                coord_res = (coord_v2 - coord_v1) / coord_bnds.shape[0]
                coord_res = float(coord_res.values)
                coord_min, coord_max = (
                    coord_v1, coord_v2) if coord_res > 0 else (
                coord_v2, coord_v1)
                dataset.attrs[coord_min_attr_name] = cast(coord_min.values)
                dataset.attrs[coord_max_attr_name] = cast(coord_max.values)
            elif coord is not None \
                    and coord.ndim == 1 \
                    and coord.shape[0] > 0:
                coord_v1 = coord[0]
                coord_v2 = coord[-1]
                if coord.shape[0] > 1:
                    coord_res = (coord_v2 - coord_v1) / (coord.shape[0] - 1)
                    coord_v1 -= coord_res / 2
                    coord_v2 += coord_res / 2
                    coord_res = float(coord_res.values)
                    coord_min, coord_max = (
                        coord_v1, coord_v2) if coord_res > 0 else (
                        coord_v2, coord_v1)
                else:
                    coord_min, coord_max = coord_v1, coord_v2
                dataset.attrs[coord_min_attr_name] = cast(coord_min.values)
                dataset.attrs[coord_max_attr_name] = cast(coord_max.values)

    dataset.attrs['date_modified'] = datetime.datetime.now().isoformat()
    return dataset


def update_dataset_var_attrs(dataset: xr.Dataset,
                             var_attrs_list: NameDictPairList) -> xr.Dataset:
    """
    Update the attributes of variables in given *dataset*.
    Optionally rename variables according to a given attribute named "name".

    *var_attrs_list* must be a sequence of pairs of the form (<var_name>, <var_attrs>) where <var_name> is a string
    and <var_attrs> is a dictionary representing the attributes to be updated , including an optional "name" attribute.
    If <var_attrs> contains an attribute "name", the variable named <var_name> will be renamed to that attribute's
    value.

    :param dataset: A dataset.
    :param var_attrs_list: List of tuples of the form (variable name, properties dictionary).
    :return: A shallow copy of *dataset* with updated / renamed variables.
    """
    if not var_attrs_list:
        return dataset

    var_name_attrs = dict()
    var_renamings = dict()
    new_var_names = set()

    # noinspection PyUnusedLocal,PyShadowingNames
    for var_name, var_attrs in var_attrs_list:
        if not var_attrs:
            continue
        # noinspection PyShadowingNames
        var_attrs = dict(var_attrs)
        if 'name' in var_attrs:
            new_var_name = var_attrs.pop('name')
            if new_var_name in new_var_names:
                raise ValueError(
                    f'variable {var_name!r} cannot be renamed into {new_var_name!r} '
                    'because the name is already in use')
            new_var_names.add(new_var_name)
            var_attrs['original_name'] = var_name
            var_renamings[var_name] = new_var_name
            var_name = new_var_name
        var_name_attrs[var_name] = var_attrs

    if var_renamings:
        dataset = dataset.rename(var_renamings)
    elif var_name_attrs:
        dataset = dataset.copy()

    if var_name_attrs:
        for var_name, var_attrs in var_name_attrs.items():
            var = dataset[var_name]
            var.attrs.update(var_attrs)

    return dataset


def update_dataset_chunk_encoding(dataset: xr.Dataset,
                                  chunk_sizes: Dict[str, int] = None,
                                  format_name: str = None,
                                  in_place: bool = False) -> xr.Dataset:
    """
    Update each variable's encoding in *dataset* with respect to *chunk_sizes*
    so *dataset* is written in chunks for given *format_name*.

    :param dataset: input dataset.
    :param chunk_sizes: the chunk sizes to be used for the encoding.
        If None, any chunking encoding is removed.
    :param format_name: format name, e.g. "zarr" or "netcdf4".
    :param in_place: If ``True``, *dataset* will be modified in place and returned.
    """
    if format_name == FORMAT_NAME_ZARR:
        chunk_sizes_attr_name = "chunks"
    elif format_name == FORMAT_NAME_NETCDF4:
        chunk_sizes_attr_name = "chunksizes"
    else:
        return dataset
    if not in_place:
        dataset = dataset.copy()
    for var_name in dataset.variables:
        var = dataset[var_name]
        if chunk_sizes is not None:
            def get_size(i):
                dim_name = var.dims[i]
                size = chunk_sizes.get(dim_name)
                if isinstance(size, int):
                    return size
                if var.chunks:
                    size = var.chunks[i]
                    if isinstance(size, int):
                        return size
                    if len(size):
                        return size[0]
                return var.shape[i]

            var.encoding.update(
                {chunk_sizes_attr_name: tuple(map(get_size, range(var.ndim)))})
        elif chunk_sizes_attr_name in var.encoding:
            # Remove any explicit and possibly unintended specification
            del var.encoding[chunk_sizes_attr_name]
    return dataset
